from typing import List, Tuple

import torch
from transformers import AutoModelForCausalLM, AutoTokenizer


class BaseModel:
    def __init__(self, model: AutoModelForCausalLM, tokenizer: AutoTokenizer, *args, **kwargs):
        self.model = model
        self.tokenizer = tokenizer


    def forward(self, input: dict) -> torch.Tensor:
        raise NotImplementedError("Subclass must implement this method")
    
    
    def generate(self, input: dict) -> str:
        raise NotImplementedError("Subclass must implement this method")
    

    def get_logits(self, input: dict) -> torch.Tensor:
        raise NotImplementedError("Subclass must implement this method")
    
    
    def get_candidates(self, input: dict, position: int) -> List[str]:
        raise NotImplementedError("Subclass must implement this method")
    

    def filter(self, prompts: List[Tuple[str, float]]) -> List[Tuple[str, float]]:
        raise NotImplementedError("Subclass must implement this method")



# import errno
# import os
# import signal
# import functools
# import ast
# import astunparse
#
#
# class TimeoutError(Exception):
#     pass
#
# def timeout(seconds=10, error_message=os.strerror(errno.ETIME)):
#     """
#     This is a simple decorator that raises a TimeoutError if the function takes more than the specified time.
#     Since some programs might end up not finishing, this is a good way to avoid infinite loops.
#     :param seconds:
#     :param error_message:
#     :return:
#     """
#     def decorator(func):
#         def _handle_timeout(signum, frame):
#             raise TimeoutError(error_message)
#
#         @functools.wraps(func)
#         def wrapper(*args, **kwargs):
#             signal.signal(signal.SIGALRM, _handle_timeout)
#             signal.alarm(seconds)
#             try:
#                 result = func(*args, **kwargs)
#             finally:
#                 signal.alarm(0)
#             return result
#
#         return wrapper
#
#     return decorator
#
# @timeout(10)
# def timeout_exec(code):
#     exec(code, globals(), globals())
#
# @timeout(10)
# def timeout_exec_with_return(code):
#     loc = {}
#     exec(code, loc, loc)
#     return loc["my_new_var"]
#
# def extract_test_code(assert_statement: str) -> str:
#     ast_parsed = ast.parse(assert_statement)
#     try:
#         call_str = ast_parsed.body[0].test.left # type: ignore
#     except:
#         call_str = ast_parsed.body[0].test # type: ignore
#
#     return astunparse.unparse(call_str).strip()
#
# def get_output_of_test(code, test):
#     test_code = extract_test_code(test)
#     failing_test = f'from typing import *\n\n{code}\n\nmy_new_var = {test_code}'
#     output = timeout_exec_with_return(failing_test)
#     return output
#



import random
import os
import guidance 

from .default_trainer import DefaultTrainer
from .utils import clean_string, _load_prompt
from .node import Node

class APOTrainer(DefaultTrainer):

    def prepare_meta_prompts(self):
        # initializer
        init_prompt = _load_prompt(os.path.join(self.args.meta_prompts_dir, "initializer.md"))
        generation_config = "n={} temperature={} max_tokens={}".format(self.args.init_n_prompts, self.args.init_temperature, self.args.prompt_max_tokens)
        prompt = init_prompt.replace("[[GENERATION_CONFIG]]", generation_config)
        self.initializer = guidance(prompt, llm=self.llm) # guidance program for initialization
        
        # inspector: ~ loss.backward()
        inspector_prompt = _load_prompt(os.path.join(self.args.meta_prompts_dir, "inspector.md"))
        self.inspector = guidance(inspector_prompt, llm=self.llm)

        # proposer: ~ optimizer.step()
        proposer_prompt = _load_prompt(os.path.join(self.args.meta_prompts_dir, "proposer.md"))
        self.proposer = guidance(proposer_prompt, llm=self.llm)
        
    def _update(self, node, batch, indices):
        failure_string = self._pack_batch(batch)

        guidance.llms.OpenAI.cache.clear()
        f = self.inspector(prompt=node.prompt,
                            failure_string=failure_string, n_reasons=5)
        gradient = clean_string(f["gradients"])

        f = self.proposer(prompt=node.prompt,
                          failure_string=failure_string, gradient=gradient, max_tokens=self.args.prompt_max_tokens)

        new_node = Node(
            timestamp=node.timestamp+1,
            id=node.id+"-{}".format(node.n_child),
            parent=node.id,
            prompt=clean_string(f["new_prompt"]),
            reasoning=gradient,
            optim_batch=indices,
        )
        node.n_child += 1

        return new_node


import os
import logging

import random
import numpy as np

from .task.collection import Task2Class
from .model.collection import Model2Class
from .trainer.collection import Trainer2Class

def get_logger(args):
    # Create a logger
    logger = logging.getLogger("LABO")
    logger.setLevel(logging.INFO)

    # Create a file handler that writes to output.log
    file_handler = logging.FileHandler(os.path.join(args.output_dir, "output.log"))
    file_handler.setLevel(logging.INFO)

    # Create a stream handler that prints to the screen
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)

    # Create a formatter for the log messages
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(filename)s - %(funcName)s - %(message)s')
    file_handler.setFormatter(formatter)
    stream_handler.setFormatter(formatter)

    # Add the handlers to the logger
    logger.addHandler(file_handler)
    logger.addHandler(stream_handler)
    
    # Avoid interference with guidance logger
    logger.propagate = False

    return logger

def get_task(args, logger):
    _cls = Task2Class(args.task) 
    task = _cls(args, logger, args.data_dir)
    task.load_data()
    return task
    
def get_model(args, logger):
    _cls = Model2Class(args.model)
    model = _cls(args, logger)
    model.load_prompt()
    return model

def get_trainer(args, logger):
    _cls = Trainer2Class(args.trainer)
    trainer = _cls(args, logger)
    return trainer

def seed_everything(seed):
	random.seed(seed)
	os.environ['PYTHONHASHSEED'] = str(seed)
	np.random.seed(seed)


import platformdirs
import re
import os
import io
from collections import Counter
from PIL import Image

from textgrad.tasks.base import Dataset
from textgrad.loss import ImageQALoss
from textgrad.variable import Variable


def compress_image(decoded_image, max_size_bytes=3.6*1024*1024):
    # First, try saving as PNG without any compression
    buffer = io.BytesIO()
    decoded_image.save(buffer, format='PNG')
    size = buffer.tell()
    
    # If the original PNG is already small enough, return it
    if size <= max_size_bytes:
        buffer.seek(0)
        return buffer.getvalue()
    
    # If PNG is too large, resize the image
    width, height = decoded_image.size
    while size > max_size_bytes:
        print(f"Compressing image to {width}x{height}...")
        width = int(width * 0.9)
        height = int(height * 0.9)
        resized_image = decoded_image.resize((width, height), Image.LANCZOS)
        
        buffer = io.BytesIO()
        resized_image.save(buffer, format='PNG')
        size = buffer.tell()
        
        if width <= 1 or height <= 1:
            raise ValueError("Unable to compress image to the desired size without excessive loss of resolution")
    
    buffer.seek(0)
    return buffer.getvalue()

def preprocess_text(text):
    # Remove punctuation and convert to lowercase
    text = re.sub(r'[^\w\s]', '', text).lower()
    return text

def find_most_similar_choice(text, choices):
    # Preprocess the given text
    text = preprocess_text(text)
    text_words = Counter(text.split())
    scores = []
    for choice in choices:
        choice_text = preprocess_text(choice)
        choice_words = Counter(choice_text.split())
        common_words = sum((text_words & choice_words).values()) # the number of common words
        scores.append(common_words)
    most_similar_index = scores.index(max(scores))
    return most_similar_index

def extract_answer(response_text):
    """
    Extract the answer from the response.
    For example, "xxxxx. Answer: (A) a gas." -> "A"
    If extraction fails, return the entire string after "Answer: ".
    """
    # Attempt to match the format "Answer: (A)" or "Answer: (a)" in case-sensitive manner
    match = re.search(r"Answer: \(([A-Z])\)", response_text)
    if match:
        return match.group(1).upper()  # Return as uppercase
    else:
        # Fallback: match the format "Answer: " followed by any characters until the next period or end of line
        fallback_match = re.search(r"Answer: ([^\.]+)", response_text)
        if fallback_match:
            return fallback_match.group(1).strip()
    return response_text

def normalize_extracted_answer(extracted_answer, question_data, options):
    # Normalize the extracted answer
    choices = question_data["choices"]
    options = options[:len(choices)]

    # 'A' -> one of the choices
    if extracted_answer in options:
        normalized_answer = options.index(extracted_answer)
        return normalized_answer

    # '(a) a gas.'
    for choice in choices:
        if choice.lower() in extracted_answer.lower():
            normalized_answer = choices.index(choice)
            return normalized_answer
        
    # find the most similar choice
    normalized_answer = find_most_similar_choice(extracted_answer, choices)
    return normalized_answer

def safe_equal(a, b):
    # Check if two intergers are equal
    return a == b

class ScienceQADataset(Dataset):
    def __init__(self, evaluation_api:str, root: str=None, split: str="test", task_instruction: str=None, evaluation_instruction: str=None, *args, **kwargs):
        """ScienceQA dataset from HF."""
        if root is None:
            root = platformdirs.user_cache_dir("textgrad")
        self.root = root
        self.split = split
        assert split in ["test"]
        self.data = self.load_scienceqa_data()
        self.options = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J"]
        self.evaluation_api = evaluation_api
        self.task_instruction = self.get_default_task_instruction(task_instruction) # NOTE: check the task instruction
        self.evaluation_instruction = self.get_default_evaluation_instruction(evaluation_instruction) # NOTE: check the evaluation instruction
        
    def __getitem__(self, index):
        row = self.data[index]
        pid = row["pid"]
        image = row["image"]
        question = row["question"]
        choices = row["choices"]
        answer = row["answer"]
        hint = row["hint"]

        query = f"{self.task_instruction}" # NOTE: Add the task description
        if hint is not None and len(hint) > 0:
            query += f"\nContext: {hint}"
        query += f"\nQuestion: {question}"
        if choices:
            choice_list = []
            for i, c in enumerate(choices):
                choice_list.append("({}) {}".format(self.options[i], c))
            choice_txt = " ".join(choice_list)
            query += f"\nChoices: {choice_txt}"

        # NOTE: convert image to bytes
        if "claude" in self.evaluation_api.model_string:
            image_bytes = compress_image(image)
            # print("Image size:", len(image_bytes))
        else:
            buffer = io.BytesIO()
            image.save(buffer, format='png')
            image_bytes = buffer.getvalue()
            buffer.close()

        # NOTE: ques_data stores other fields that might be useful later
        ques_data = {
            "pid": pid,
            "question": question,
            "choices": choices,
            "hint": hint
        }
        test_time_objective = self._get_instance_test_time_objective(query, image_bytes)
        instance_eval_fn = self._get_instance_eval_fn(query, answer, ques_data)
        return image_bytes, query, answer, ques_data, test_time_objective, instance_eval_fn # NOTE: check the sample format

    def __len__(self):
        return len(self.data)

    def get_default_task_instruction(self, instruction):
        if instruction is not None:
            print("Using user-defined task instruction:\n", instruction, "\n")
            task_instruction = instruction
        else:
            task_instruction = "You will answer a scientific question based on an image. Please ensure you accurately interpret the image and think step by step. The last line of your answer should be formatted as follows: 'Answer: (X) Your Option.'"
        return task_instruction
    
    def load_scienceqa_data(self):
        scienceqa_dir = os.path.join(self.root, "scienceqa")
        try:
            from datasets import Dataset
            data = Dataset.load_from_disk(scienceqa_dir)
            print("Loaded ScienceQA dataset from cache.")
            return data
        except FileNotFoundError:
            from datasets import load_dataset
            data = load_dataset("derek-thomas/ScienceQA", split=self.split)
            data_img = data.filter(lambda x: x['image'] is not None) # select examples with a non-empty question image
            data_img = data_img.map(lambda x, i: {'pid': str(i), **x}, with_indices=True) # add index ID (string) for each example
            data_img.save_to_disk(scienceqa_dir)
            print("Loaded ScienceQA dataset from HF.")
            return data_img
        
    def get_default_evaluation_instruction(self, instruction):
        if instruction is not None:
            print("Using user-defined evaluation instruction:\n", instruction, "\n")
            evaluation_instruction = instruction
        else:
            evaluation_instruction = "Please evaluate the existing answer to the visual scientific problem without solving it yourself. Verify that the answer accurately understands the image, provides appropriate knowledge and reasoning logic to address the question."
        return evaluation_instruction

    def _get_instance_test_time_objective(self, question: str, image: bytes):
        """Define the loss function for the test time optimization."""
        eval_fn = ImageQALoss(evaluation_instruction=self.evaluation_instruction, engine=self.evaluation_api)
        def test_time_objective(instance: Variable):
            var_image = Variable(image, role_description="image input", requires_grad=False)
            var_question = Variable(question, role_description="question input", requires_grad=False)
            return eval_fn(question=var_question, image=var_image, response=instance)
        return test_time_objective
        
    def eval_extraction_and_matching(self, response_text, correct_answer, question_data):
        # Extract the precited answer text from the response
        extracted_answer  = extract_answer(response_text)

        # Normalize the extracted answer to match the answer type
        normalized_answer = normalize_extracted_answer(extracted_answer, question_data, self.options)

        # Verify the prediction is true or false
        true_false = safe_equal(normalized_answer, correct_answer)

        # Calculate the score and store the result data
        # NOTE: check the result data format
        score = 1 if true_false else 0
        result_data = {
            "extracted_answer": extracted_answer,
            "normalized_answer": normalized_answer,
            "true_false": true_false
        }
        return score, result_data

    def _get_instance_eval_fn(self, question_prompt: str, answer: str, ques_data: dict):
        """
        Define the evaluation function for scoring the response.
        Extraxct the short answer from the response and compare it with the ground truth.
        """
        # NOTE: check the evaluation function format
        eval_extraction_based_fn = lambda response: self.eval_extraction_and_matching(response.value, answer, ques_data)
        return eval_extraction_based_fn



import os

from .trainer.default_trainer import DefaultTrainer
from .trainer.node import Node

from .utils import get_task, get_model, get_trainer

def run(args, logger):
    
    # load data
    task = get_task(args, logger)
    logger.info("Initializing task: {}".format(task))

    # load model
    model = get_model(args, logger)

    # get trainner
    trainer = get_trainer(args, logger)
    initial_nodes = trainer.initialize(task)
    
    if args.do_train:
        best_score, best_node, states = trainer.train(model, initial_nodes, task)
    
    if args.do_test:
        if args.do_train:
            test_node = best_node
        else:
            assert len(initial_nodes)==1
            test_node = initial_nodes[0]
        
        data = task.get_data_split("test")
        result, final_score = trainer.evaluate(model, test_node, task, data, "test")
        
        logger.info("[Test] Node ID: {}; Score: {}".format(test_node.id, final_score))
        
        # result.to_csv(os.path.join(args.output_dir, "test_result.csv"))

    return best_node.prompt



from .dataloader import GreaterDataloader
from .optimizer import ApeOptimizer, ApoOptimizer, GreaterOptimizer, Pe2Optimizer, TextGradOptimizer


__all__ = [
    "GreaterDataloader",
    "ApeOptimizer", "ApoOptimizer", "GreaterOptimizer", "Pe2Optimizer", "TextGradOptimizer"
]



import os
import re
import json

import numpy as np
import pandas as pd

import random
import string

from .task import AbstractTask
from .ii_utils import get_em_score

class ZeroshotCoTTask(AbstractTask):
    def _load_split(self, split):
        filename = os.path.join(self.data_dir, "{}.csv".format(split))
        data, size = None, 0
        
        if os.path.exists(filename):
            data = pd.read_csv(filename, index_col=0)
            size = len(data)

        if self.args.debug:
            data = data.sample(n=20, random_state=self.args.seed)
        
        return data, size
    
    def load_data(self):
        self.train_data, self.train_size = self._load_split("train")
        self.dev_data, self.dev_size = self._load_split("dev")
        self.test_data, self.test_size = self._load_split("test")
        
        self.data["train"] = self.train_data
        self.data["dev"] = self.dev_data
        self.data["test"] = self.test_data

        self.logger.info("Loading data... # Train: {}; # Dev: {}; # Test: {}".format(self.train_size, self.dev_size, self.test_size))
        
    def evaluate(self, result_df):
        result_df = self._postprocess(result_df)
        labels = result_df["label"].tolist()
        outputs = result_df["output"].tolist()

        scores = []
        for prediction, ans_ in zip(outputs, labels):
            if self.args.task == "math":
                score = prediction == float(ans_)
            elif self.args.task == "bbh":
                score = get_em_score(prediction, str(ans_))
            scores.append(score)

        result_df["score"] = scores
        return result_df, result_df["score"].mean()

    def _postprocess(self, data):
        all_reasoning = []
        all_output = []
        for pred in data["raw_output"].tolist():
            final_pred = pred.strip(" ").strip(".").strip('"').strip("'").strip(":").strip(" ")
            all_reasoning.append(pred)
            all_output.append(final_pred)

        data["output"] = all_output

        return data


import hashlib
import time

try:
    import leetcode
    import leetcode.auth
    from leetcode_env.types import LeetCodeSubmission, ProgrammingLanguage
except:
    raise Exception("Please install leetcodehard gym (https://github.com/GammaTauAI/leetcode-hard-gym)")

from textgrad.engine import get_engine
from diskcache import Cache
import os
from dotenv import load_dotenv

# if black is missing generate exception

try:
    from black import format_str, FileMode
    import autopep8
except:
    raise Exception("Please install black and autopep8")

import ast


class LeetCodeEvaluator:
    """
    Wrapper on top of the leetcode hard gym package.
    Note that this needs an environment variable, namely LEETCODE_SESSION to be set.
    In addition to this, the leetcode-cli package has to be manually edited:

    See the following issue on GitHub:
    https://github.com/GammaTauAI/leetcode-hard-gym/issues/20

    :return:
    """

    def __init__(self):
        configuration = leetcode.Configuration()
        leetcode_session = os.environ["LEETCODE_SESSION"]
        csrf_token = os.environ["LEETCODE_CSRF_TOKEN"]

        configuration.api_key["x-csrftoken"] = csrf_token
        configuration.api_key["csrftoken"] = csrf_token
        configuration.api_key["LEETCODE_SESSION"] = leetcode_session
        configuration.api_key["Referer"] = "https://leetcode.com"
        configuration.debug = False
        self.leet_code_cache = Cache("./cache_leetcode")

        self.api_instance = leetcode.DefaultApi(leetcode.ApiClient(configuration))

    def id_from_slug(self, slug: str) -> str:
        """
        Retrieves the id of the question with the given slug
        """
        graphql_request = leetcode.GraphqlQuery(
            query="""
                      query getQuestionDetail($titleSlug: String!) {
                        question(titleSlug: $titleSlug) {
                          questionId
                        }
                      }
                  """,
            variables={"titleSlug": slug},
            operation_name="getQuestionDetail",
        )
        response = ast.literal_eval(str(self.api_instance.graphql_post(body=graphql_request)))
        frontend_id = response['data']['question']['question_id']
        return frontend_id

    def submit_for_evaluation(self, code, problem_tag):
        from leetcode_env.utils.formatting import PythonSubmissionFormatter

        # need to go from humaneval to leetcode and then rsstrip newlines
        code = PythonSubmissionFormatter.to_leetcode(code).rstrip().lstrip()

        # format the code
        code = format_str(code, mode=FileMode())
        code = autopep8.fix_code(code)

        # Some programs are not well formatted for the leetcode API
        TEST_ENGINE = "gpt-4o"
        ENGINE_API = get_engine(engine_name=TEST_ENGINE)
        FIXING_CODE_PROMPT = """Remove the unnecessary parentheses from the code. DO NOT CHANGE THE CODE.
Use a Python code block to write your response. For example:
```python
print('Hello world!')
```
"""
        code = ENGINE_API.generate(code, system_prompt=FIXING_CODE_PROMPT)
        code = code.split("```python")[1].split("```")[0]

        sub = LeetCodeSubmission(code=code,
                                 lang=ProgrammingLanguage.PYTHON3,
                                 question_slug=problem_tag,
                                 timeout=10)

        sub.question_id = self.id_from_slug(problem_tag)

        submission = leetcode.Submission(
            judge_type="large",
            typed_code=code,
            question_id=sub.question_id,
            test_mode=False,
            lang=sub.lang.value,
        )

        try:

            # we submit a program and then we wait for the submission to be processed
            submission_id = self.api_instance.problems_problem_submit_post(
                problem=problem_tag, body=submission)
            time.sleep(10)

        except Exception as e:
            print("Error in submission", problem_tag)
            print(e)
            print("***" * 10)
            print(code)
            print("***" * 10)
            return {"status_msg": False, "total_correct": 0, "total_testcases": 0}

        time.sleep(5)

        submission_result = self.api_instance.submissions_detail_id_check_get(
            id=submission_id.submission_id
        )

        return submission_result

    def check_if_in_cache_or_submit(self, task_id, code):
        key = hashlib.sha256(f"{task_id}__{code}".encode()).hexdigest()
        key_run_success = f"{key}_run_success"
        key_total_correct = f"{key}_total_correct"
        key_total_tests = f"{key}_total_tests"
        key_runtime = f"{key}_runtime"

        if key_run_success in self.leet_code_cache:
            success_or_error = self.leet_code_cache[key_run_success]
            total_correct = self.leet_code_cache[key_total_correct]
            total_tests = self.leet_code_cache[key_total_tests]
            runtime_res = self.leet_code_cache[key_runtime]
            print(f"Task ID: {task_id} - Success: {success_or_error} - Correct: {total_correct}/{total_tests}")

        else:
            response = self.submit_for_evaluation(code, task_id)

            if "status_msg" not in response:
                print(response)
                self.leet_code_cache[key_run_success] = False
                self.leet_code_cache[key_total_correct] = 0
                self.leet_code_cache[key_total_tests] = 0
                self.leet_code_cache[key_runtime] = -1
                return False, 0, 0, -1

            if response["status_msg"] == "Accepted":
                success_or_error = True
            else:
                success_or_error = False

            total_correct = response["total_correct"]
            total_tests = response["total_testcases"]
            runtime_res = response["status_runtime"]

            self.leet_code_cache[key_run_success] = success_or_error
            self.leet_code_cache[key_total_correct] = total_correct
            self.leet_code_cache[key_total_tests] = total_tests
            self.leet_code_cache[key_runtime] = runtime_res

            print(f"Task ID: {task_id} - Success: {success_or_error} - Correct: {total_correct}/{total_tests}")
        return success_or_error, total_correct, total_tests, runtime_res




import re
import os
import json
import argparse
import concurrent
from tqdm import tqdm
import numpy as np
from dotenv import load_dotenv
load_dotenv(override=True)
from collections import Counter

import textgrad as tg
from textgrad.tasks.multimodal import load_multimodal_instance_task

def load_checkpoint_results(checkpoint_file):
    all_solutions = {}
    if checkpoint_file.endswith(".jsonl"):
        with open(checkpoint_file, "r") as f:
            for line in f:
                data = json.loads(line)
                all_solutions[data["ques_data"]["pid"]] = data
    elif checkpoint_file.endswith(".json"):
        with open(checkpoint_file, "r") as f:
            all_solutions = json.load(f)
    print("Loaded existing results from", checkpoint_file)
    print(f"There are {len(all_solutions)} samples in the file!")
    return all_solutions

class MajorityVoting:
    def __init__(self, solutions, max_iterations):
        self.solutions = solutions
        self.max_iterations = max_iterations

    def majority_element(self, predictions):
        count = Counter(predictions)
        return max(count.keys(), key=count.get)

    def compute_accuracy(self):
        final_score = 0
        for n in range(1, self.max_iterations + 1):
            correct = 0
            for _, solution in self.solutions.items():
                answer = solution["answer"]
                predictions = [res["normalized_answer"] for res in solution["result_data"]]
                predictions = predictions[1:n+1]
                majority = self.majority_element(predictions)
                if majority == answer:
                    correct += 1
            acc = round(correct / len(self.solutions), 3)
            if acc > final_score:
                final_score = acc
        return final_score

def config():
    parser = argparse.ArgumentParser(description="Optimize a prompt for a task.")
    parser.add_argument("--task", type=str, default="mathvista", choices = ["mathvista", "scienceqa"], help="The task to evaluate the model on.")
    parser.add_argument("--multimodal", action='store_true', help="Determine if we want to load multimodal dataset.")
    parser.add_argument("--task_instruction", type=str, default=None, help="The instruction for the task.")
    parser.add_argument("--evaluation_instruction", type=str, default=None, help="The instruction for the evaluation.")
    parser.add_argument("--instance_role", type=str, default=None, help="The role description of the instance.")
    parser.add_argument("--image_role", type=str, default=None, help="The role description of the image.")
    parser.add_argument("--question_role", type=str, default=None, help="The role description of the question.")
    parser.add_argument("--cache_root", type=str, default=None, help="The cache directory to save data.")
    parser.add_argument("--results_dir", type=str, default="./results/solution_optimization_mm", help="The directory to save the results.")
    parser.add_argument("--output_file", type=str, default="", help="The output file to save or load the results.")
    parser.add_argument("--save_every", type=int, default=50, help="The frequency to save the results.")
    parser.add_argument("--engine", type=str, default="gpt-4o", help="The API to use for inference.")
    parser.add_argument("--eval_engine", type=str, default="gpt-4o", help="The API to use for evaluation.")
    parser.add_argument("--max_iterations", type=int, default=3, help="The maximum number of iterations of test-time updates.")
    parser.add_argument("--num_threads", type=int, default=16, help="The number of threads to use for evaluation.")
    parser.add_argument("--test_num", type=int, default=-1, help="The number of test samples to evaluate., -1 means all.")
    parser.add_argument("--majority_voting", action='store_true', default=False, help="Determine if we want to use majority voting.")
    return parser.parse_args()

def get_zeroshot_answer(question, mm_data=None):
    """Getting the zero-shot answer from an LLM without optimizing the response at test time."""
    # The system prompt is from: https://github.com/openai/simple-evals/blob/main/sampler/chat_completion_sampler.py
    STARTING_SYSTEM_PROMPT = (
        "You are ChatGPT, a large language model trained by OpenAI, based on the GPT-4 architecture."
        + "\nKnowledge cutoff: 2023-12\nCurrent date: 2024-07-01"
    )
    if args.question_role is not None:
        question_role = args.question_role
    else:
        question_role = "question to the language model"
    if args.image_role is not None:
        image_role = args.image_role
    else:
        image_role = "image associate with question to the language model"
    system_prompt = tg.Variable(STARTING_SYSTEM_PROMPT, requires_grad=False, role_description="system prompt to the language model")
    if mm_data is None: # vanilla text.
        model = tg.BlackboxLLM(llm_engine, system_prompt)
        response = model(tg.Variable(question, requires_grad=False, role_description=question_role))
    else: # multi-modal (such as image).
        from textgrad.autograd import MultimodalLLMCall
        variable_mm_data = tg.Variable(mm_data, requires_grad=False, role_description=image_role)
        variable_question = tg.Variable(question, requires_grad=False, role_description=question_role)
        response = MultimodalLLMCall(args.engine)([variable_mm_data, variable_question])
    return response

def run_test_time_training(sample):
    performance_history = []
    results_data = []
    mm_data, question, answer, ques_data, test_time_objective, instance_eval_fn = sample # NOTE: check the sample format
    zero_shot_response = get_zeroshot_answer(question, mm_data)

    # NOTE: Set the instance role description
    if args.instance_role is not None:
        instance_role_description = args.instance_role
    else:
        if args.task == "mathvista":
            instance_role_description = "detailed reasoning and prediction for the image-related math question."
        elif args.task == "scienceqa":
            instance_role_description = "detailed reasoning and prediction for the image-related science question."
        else:
            instance_role_description = "precise and short answer to the image-related question."

    instance_var = tg.Variable(zero_shot_response.value, requires_grad=True, role_description=instance_role_description)
    
    # Evaluate the zero-shot response
    # NOTE: The instance_eval_fn should return a tuple of (score, result_data) if result_data is needed.
    eval_score = instance_eval_fn(instance_var)
    if isinstance(eval_score, tuple):
        score, result_data = eval_score
    else:
        score = eval_score
        result_data = None
    performance_history.append(int(score))
    results_data.append(result_data)
    
    optimizer = tg.TextualGradientDescent(engine=llm_engine, 
                                        parameters=[instance_var], 
                                        constraints=["Your response should answer the question given an image."])

    predictions = []
    predictions.append(tg.Variable(
        instance_var.value,
        role_description=instance_var.role_description
        ))
    
    loss_history = []
    
    # Start test time optimization
    for _ in range(args.max_iterations):
        optimizer.zero_grad()
        # Compute the test time loss
        test_time_loss = test_time_objective(instance_var)
        loss_history.append(str(test_time_loss))
        test_time_loss.backward()
        optimizer.step()

        eval_score = instance_eval_fn(instance_var)
        # NOTE: The instance_eval_fn should return a tuple of (score, result_data) if result_data is needed.
        if isinstance(eval_score, tuple):
            score, result_data = eval_score
        else:
            score = eval_score
            result_data = None

        predictions.append(tg.Variable(
            instance_var.value,
            role_description=instance_var.role_description
        ))
        performance_history.append(int(score))
        results_data.append(result_data)

    return question, answer, predictions, performance_history, loss_history, results_data, ques_data # NOTE: check the return format

args = config()
llm_engine = tg.get_engine(engine_name=args.engine)
tg.set_backward_engine(llm_engine, override=True)

eval_engine = tg.get_engine(engine_name=args.eval_engine)
test_set = load_multimodal_instance_task(args.task,
                                         evaluation_api=eval_engine, 
                                         task_instruction=args.task_instruction, # NOTE: check the instruction
                                         evaluation_instruction=args.evaluation_instruction, # NOTE: check the instruction
                                         root=args.cache_root)

# Create the results directory if it does not exist
os.makedirs(args.results_dir, exist_ok=True)
output_file = f"{args.task}_predictions.json" if args.output_file == "" else args.output_file
output_file = os.path.join(args.results_dir, output_file)
output_jsonl_file = output_file.replace(".json", ".jsonl")

# Read the checkpoint result file if it exists
if os.path.exists(output_jsonl_file):
    all_solutions = load_checkpoint_results(output_jsonl_file)
elif os.path.exists(output_file):
    all_solutions = load_checkpoint_results(output_file)
else:
    all_solutions = {}
all_history = [solution["performance_history"] for solution in all_solutions.values()]

if args.num_threads > 0:
    # Use ThreadPoolExecutor
    with concurrent.futures.ThreadPoolExecutor(max_workers=args.num_threads) as executor:
        futures = []
        print("\nPreparing to run test-time training...")
        for i, sample in enumerate(test_set):
            image_bytes, query, answer, ques_data, test_time_objective, instance_eval_fn = sample # NOTE: check the sample format
            pid = ques_data["pid"] # NOTE: check the question id
            if pid in all_solutions:
                # print(f"Skipping sample {pid}")
                continue
            if args.test_num > 0 and i >= args.test_num:
                break
            future = executor.submit(run_test_time_training, sample)
            futures.append(future)

        print("\nRunning test-time training...")
        for _, future in enumerate(tqdm(concurrent.futures.as_completed(futures), total=len(futures), position=0)):
            question, answer, predictions, performance_history, loss_history, result_data, ques_data = future.result() # NOTE: check the return format
            pid = ques_data["pid"] # NOTE: check the question id
            all_solutions[pid] = {
                "question": question,
                "answer": answer, 
                "predictions": [p.value for p in predictions], 
                "loss_history": loss_history,
                "performance_history": performance_history,
                "result_data": result_data,
                "ques_data": ques_data
                }
            all_history.append(performance_history)
            # save result to output file in append mode
            with open(output_jsonl_file, "a") as f:
                json_line = json.dumps(all_solutions[pid])
                f.write(json_line + "\n")
else:
    # Use regular for loop
    if args.test_num > 0:
        total = min(args.test_num, len(test_set))
    else:
        total = len(test_set)
    for i, sample in enumerate(tqdm(test_set, total=total, position=0)):
        image_bytes, query, answer, ques_data, test_time_objective, instance_eval_fn = sample # NOTE: check the sample format
        pid = ques_data["pid"] # NOTE: check the question id
        if pid in all_solutions:
            continue
        if args.test_num > 0 and i >= args.test_num:
            break
        future = run_test_time_training(sample)

        question, answer, predictions, performance_history, loss_history, result_data,  ques_data = future # NOTE: check the return format
        result = {
            "question": question,
            "answer": answer,
            "predictions": [p.value for p in predictions],
            "performance_history": performance_history,
            "loss_history": loss_history,
            "result_data": result_data,
            "ques_data": ques_data
            }
        all_solutions[pid] = result
        all_history.append(performance_history)

        # save result to output file in append mode
        with open(output_jsonl_file, "a") as f:
            json_line = json.dumps(result)
            f.write(json_line + "\n")
        if i % args.save_every == 0:
            with open(output_file, "w") as f:
                print("Saving results to", output_file)
                json.dump(all_solutions, f, indent=4)

# Quick summary of the results
all_solutions = {k: all_solutions[k] for k in sorted(all_solutions, key=lambda x: int(x))} # sorted by int(pid)
with open(output_file, "w") as f:
    json.dump(all_solutions, f, indent=4)

print("\nSummary of the results:")
score_history = np.array(all_history).mean(axis=0)
score_history = [round(float(s), 3) for s in score_history]
print(score_history)

if args.majority_voting:
    print("\nMajority Voting:")
    majority_voting = MajorityVoting(all_solutions, args.max_iterations)
    final_score = majority_voting.compute_accuracy()
else:
    final_score = round(max(np.array(all_history).mean(axis=0)), 3)

print(f"Final Accuracy: {final_score}")
print(f"Total samples: {len(all_solutions)}")
print("Done!")



import sys
sys.path.append("../src")
import json
from greaterprompt.optimizer import GreaterOptimizer, GreaterDataloader

import torch
import streamlit as st
from torch.nn import functional as F
from transformers import AutoModelForCausalLM, AutoTokenizer


CUDA_DEVICES = torch.cuda.device_count()

with st.sidebar:
    settings = st.markdown("<h3>🛠️ Advanced Settings</h3>", unsafe_allow_html=True)
    device = st.selectbox("Device", ["cpu", *[f"cuda:{i}" for i in range(CUDA_DEVICES)]], index=1)
    intersect_q = st.number_input("Intersect Q", value=5)
    candidates_topk = st.number_input("Candidates Topk", value=10)
    perplexity_loss = st.selectbox("Perplexity Loss", [True, False])
    perplexity_lambda = st.number_input("Perplexity Lambda", value=0.2)
    filter = st.selectbox("Filter", [True, False])
    rounds = st.number_input("Rounds", value=80)

st.markdown("<h1 style='text-align: center; white-space: nowrap;'>🤩 Optimize with GreaterOptimizer</h1>", unsafe_allow_html=True)
model_path = st.text_input("Model Path", key="model_path", value='google/gemma-2-9b-it')
uploaded_file = st.file_uploader("Upload a jsonl input file", type=("jsonl"))

p_extraction = st.text_input("P Extractor", value="\nNext, only give the exact answer, no extract words or any punctuation:")

col1, col2, col3 = st.columns([1, 2, 1])
with col2:
    start_button = st.button("⚡ Start Optimization", type="primary", use_container_width=True)

optimize_config = {
    "intersect_q": intersect_q,
    "candidates_topk": candidates_topk,
    "loss_function": F.cross_entropy,
    "perplexity_loss": perplexity_loss,
    "perplexity_lambda": perplexity_lambda,
    "filter": filter,
    "generate_config": {
        "do_sample": True,
        "temperature": 0.2,
        "max_new_tokens": 512
    }
}

if not uploaded_file and start_button:
    st.info("Please upload a jsonl file to start")

if p_extraction.strip() == "" and start_button:
    st.info("Please enter a valid P Extractor to start")

if uploaded_file and p_extraction.strip() and start_button:
    content = uploaded_file.getvalue().decode("utf-8")
    lines = content.strip().split('\n')

    inputs = []
    
    for line in lines:
        if line.strip():
            try:
                inputs.append(json.loads(line))
            except json.JSONDecodeError:
                st.error(f"Invalid JSON in line: {line}")

    with st.status(f"Loading model from {model_path}...", expanded=True) as status:
        st.write(f"Loading model to {device}...")
        model = AutoModelForCausalLM.from_pretrained(model_path, torch_dtype=torch.bfloat16, device_map=device)

        st.write("Enabling gradient checkpointing...")
        model.gradient_checkpointing_enable()
        
        st.write("Loading tokenizer...")
        tokenizer = AutoTokenizer.from_pretrained(model_path)

        st.write("Loading optimizer...")
        optimizer = GreaterOptimizer(model=model, tokenizer=tokenizer, optimize_config=optimize_config)

        st.write("Building data inputs...")
        dataset = GreaterDataloader(custom_inputs=inputs)
        
        status.update(label="🥰 Model and Optimizer loaded successfully! Everything is ready now, start optimizing!", state="complete")

    with st.status("⚡ Optimizing...", expanded=True) as status:
        progress_bar = st.progress(0)
        status_info = st.empty()
        
        def update_progress(progress, info):
            progress_bar.progress(progress)
            
            if "status" in info and info["status"] == "complete":
                status_info.text("Optimization completed! Here are the results:")
            else:
                batch_info = f"Batch {info['batch']}/{info['total_batches']}"
                round_info = f"Round {info['round']}/{info['total_rounds']}"
                status_info.text(f"{batch_info} - {round_info}")
        
        outputs = optimizer.optimize_streamlit(inputs=dataset, p_extractor=p_extraction, rounds=rounds, callback=update_progress)
        
        status.update(label="😄 Optimization complete!", state="complete")
        st.markdown("<h5 style='text-align: left; white-space: nowrap;'>🎯 Optimization Result:</h5>", unsafe_allow_html=True)
        st.write(outputs)



from .functional import sum, aggregate
from .llm_ops import LLMCall, FormattedLLMCall, LLMCall_with_in_context_examples
from .multimodal_ops import MultimodalLLMCall, OrderedFieldsMultimodalLLMCall
from .function import Module
from .string_based_ops import StringBasedFunction


import streamlit as st


p2p_star = {
    "not ( True ) and ( True ) is": "Use this statement with a conditional if know what is the value True of and what Not False means. Or not True and also boolean. In explain your reasoning.",
    "((-1 + 2 + 9 * 5) - (-2 + -4 + -4 * -7)) =": "Use your knowledge reasoning and think step by step. Finally give the actual correct answer.",
    "Today is Christmas Eve of 1937. What is the date tomorrow in MM/DD/YYYY?": "Use the date today which will not would give us an error. solution is given as answer date is correct the option data and the current month and year to get to previous and current month of year to determine what the current data will look."
}

st.set_page_config(
    layout="wide",
    page_title="GreaterOptimizer",
    initial_sidebar_state="expanded"
)

st.markdown("""
<style>
    header {
        visibility: hidden;
    }

    .block-container {
        padding-top: 2rem !important;
        padding-bottom: 0rem;
    }
    
    .main-header {
        display: flex;
        justify-content: center;
        text-align: center;
        padding: 0;
        margin: 0 auto;
        max-width: 100%;
    }
    .title-text {
        font-size: 2.5rem;
        font-weight: bold;
        margin-top: 0px;
        margin-bottom: 0px;
        white-space: nowrap;
    }
    .author-text {
        font-size: 1.1rem;
        margin-top: 0;
        margin-bottom: 0px;
        white-space: nowrap;
    }
    .author-text a {
        text-decoration: none;
        color: inherit;
    }
    .author-text a:hover {
        text-decoration: underline;
        color: #4682B4;
    }
    .button-container {
        display: flex;
        justify-content: center;
        gap: 100px;
        margin-top: 10px;
    }
    .button-item {
        display: flex;
        flex-direction: column;
        align-items: center;
        text-align: center;
    }
    .button-text {
        margin-top: 10px;
        color: #4682B4;
        font-weight: bold;
        font-size: 18px;
    }
    
    footer {
        visibility: hidden;
    }
    
    .overview-text {
        margin-top: 20px;
        text-align: justify;
        max-width: 1000px;
        margin-left: auto;
        margin-right: auto;
        line-height: 1.6;
    }
</style>
""", unsafe_allow_html=True)

st.markdown('<div class="main-header"><h1 class="title-text">GreaterOptimizer: A Python Toolkit for Prompt Optimization</h1></div>', unsafe_allow_html=True)
st.markdown('<div class="main-header"><h3 class="author-text">\
            <a href="mailto:wmz5132@psu.edu">Wenliang Zheng</a>,\
            <a href="mailto:sfd5525@psu.edu">Sarkar Snigdha Sarathi Das</a>,\
            <a href="mailto:rmz5227@psu.edu">Rui Zhang</a>\
            </h3></div>', unsafe_allow_html=True)

st.markdown("""
<div class="button-container">
    <div class="button-item">
        <a href="https://arxiv.org/pdf/2412.09722" target="_blank" style="text-decoration: none;">
            <img src="https://upload.wikimedia.org/wikipedia/commons/8/87/PDF_file_icon.svg" width="80" height="80">
            <div class="button-text">Paper</div>
        </a>
    </div>
    <div class="button-item">
        <a href="https://github.com/WenliangZhoushan/GreaterPrompt" target="_blank" style="text-decoration: none;">
            <img src="https://cdn-icons-png.flaticon.com/512/25/25231.png" width="80" height="80">
            <div class="button-text">Code and Data</div>
        </a>
    </div>
</div>
""", unsafe_allow_html=True)

st.markdown("""
<div class="overview-text">
<p><strong>Overview of Research</strong>: The performance of large language models (LLMs) is significantly influenced by prompt design, making prompt optimization a crucial area of study. Traditional methods for optimizing prompts heavily depend on textual feedback from large, closed-source models like GPT-4, which analyze inference errors and suggest refinements. However, this reliance on computationally expensive LLMs limits the efficiency of smaller, open-source models that lack the ability to generate high-quality optimization feedback on their own.</p>

<p>This research introduces GReaTer, a novel prompt optimization technique that leverages gradient information over reasoning to enhance prompt effectiveness for smaller LLMs without relying on external, proprietary models. Unlike prior approaches that operate purely in the text space, GReaTer utilizes task loss gradients, allowing direct optimization of prompts. This method empowers smaller, open-source models to achieve state-of-the-art performance without assistance from larger LLMs.</p>
</div>
""", unsafe_allow_html=True)

st.markdown("<br>", unsafe_allow_html=True)

with st.sidebar:
    see = st.markdown("<h3>👀 See Examples</h3>", unsafe_allow_html=True)
    questions = st.selectbox("Select a question", [
        "not ( True ) and ( True ) is", "((-1 + 2 + 9 * 5) - (-2 + -4 + -4 * -7)) =", "Today is Christmas Eve of 1937. What is the date tomorrow in MM/DD/YYYY?"
    ])
    p_init = st.text_input(label="Initial P", value="Use logical reasoning to think it.", disabled=True)
    p_star = st.text_area(label="Optimized P*", value=p2p_star[questions], height=200, disabled=True)

st.write("<h2 style='text-align: center; white-space: nowrap;'>🤗 Now pick a method on the left side bar to get started!</h2>", unsafe_allow_html=True)



# try:
#     from openai import OpenAI, AzureOpenAI
# except ImportError:
#     raise ImportError("If you'd like to use OpenAI models, please install the openai package by running `pip install openai`, and add 'OPENAI_API_KEY' to your environment variables.")

import os
import json
import base64
import platformdirs
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
)
from typing import List, Union

from .base import EngineLM, CachedEngine
from .engine_utils import get_image_type_from_bytes

# Default base URL for OLLAMA
OLLAMA_BASE_URL = 'http://localhost:11434/v1'

# Check if the user set the OLLAMA_BASE_URL environment variable
if os.getenv("OLLAMA_BASE_URL"):
    OLLAMA_BASE_URL = os.getenv("OLLAMA_BASE_URL")

class ChatOpenAI(EngineLM, CachedEngine):
    DEFAULT_SYSTEM_PROMPT = "You are a helpful, creative, and smart assistant."

    def __init__(
        self,
        model_string: str="gpt-3.5-turbo-0613",
        system_prompt: str=DEFAULT_SYSTEM_PROMPT,
        is_multimodal: bool=False,
        base_url: str=None,
        **kwargs):
        """
        :param model_string:
        :param system_prompt:
        :param base_url: Used to support Ollama
        """
        root = platformdirs.user_cache_dir("textgrad")
        cache_path = os.path.join(root, f"cache_openai_{model_string}.db")

        super().__init__(cache_path=cache_path)

        self.system_prompt = system_prompt
        self.base_url = base_url
        
        if not base_url:
            if os.getenv("OPENAI_API_KEY") is None:
                raise ValueError("Please set the OPENAI_API_KEY environment variable if you'd like to use OpenAI models.")
            
            # self.client = OpenAI(
            #     api_key=os.getenv("OPENAI_API_KEY")
            # )
        elif base_url and base_url == OLLAMA_BASE_URL:
            pass
            # self.client = OpenAI(
            #     base_url=base_url,
            #     api_key="ollama"
            # )
        else:
            raise ValueError("Invalid base URL provided. Please use the default OLLAMA base URL or None.")

        self.model_string = model_string
        self.is_multimodal = is_multimodal

    @retry(wait=wait_random_exponential(min=1, max=5), stop=stop_after_attempt(5))
    def generate(self, content: Union[str, List[Union[str, bytes]]], system_prompt: str=None, **kwargs):
        if isinstance(content, str):
            return self._generate_from_single_prompt(content, system_prompt=system_prompt, **kwargs)
        
        elif isinstance(content, list):
            has_multimodal_input = any(isinstance(item, bytes) for item in content)
            if (has_multimodal_input) and (not self.is_multimodal):
                raise NotImplementedError("Multimodal generation is only supported for Claude-3 and beyond.")
            
            return self._generate_from_multiple_input(content, system_prompt=system_prompt, **kwargs)

    def _generate_from_single_prompt(
        self, prompt: str, system_prompt: str=None, temperature=0, max_tokens=2000, top_p=0.99
    ):

        sys_prompt_arg = system_prompt if system_prompt else self.system_prompt

        cache_or_none = self._check_cache(sys_prompt_arg + prompt)
        if cache_or_none is not None:
            return cache_or_none

        response = self.client.chat.completions.create(
            model=self.model_string,
            messages=[
                {"role": "system", "content": sys_prompt_arg},
                {"role": "user", "content": prompt},
            ],
            frequency_penalty=0,
            presence_penalty=0,
            stop=None,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
        )

        response = response.choices[0].message.content
        self._save_cache(sys_prompt_arg + prompt, response)
        return response

    def __call__(self, prompt, **kwargs):
        return self.generate(prompt, **kwargs)

    def _format_content(self, content: List[Union[str, bytes]]) -> List[dict]:
        """Helper function to format a list of strings and bytes into a list of dictionaries to pass as messages to the API.
        """
        formatted_content = []
        for item in content:
            if isinstance(item, bytes):
                # For now, bytes are assumed to be images
                image_type = get_image_type_from_bytes(item)
                base64_image = base64.b64encode(item).decode('utf-8')
                formatted_content.append({
                    "type": "image_url",
                    "image_url": {
                        "url": f"data:image/{image_type};base64,{base64_image}"
                    }
                })
            elif isinstance(item, str):
                formatted_content.append({
                    "type": "text",
                    "text": item
                })
            else:
                raise ValueError(f"Unsupported input type: {type(item)}")
        return formatted_content

    def _generate_from_multiple_input(
        self, content: List[Union[str, bytes]], system_prompt=None, temperature=0, max_tokens=2000, top_p=0.99
    ):
        sys_prompt_arg = system_prompt if system_prompt else self.system_prompt
        formatted_content = self._format_content(content)

        cache_key = sys_prompt_arg + json.dumps(formatted_content)
        cache_or_none = self._check_cache(cache_key)
        if cache_or_none is not None:
            return cache_or_none

        response = self.client.chat.completions.create(
            model=self.model_string,
            messages=[
                {"role": "system", "content": sys_prompt_arg},
                {"role": "user", "content": formatted_content},
            ],
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
        )

        response_text = response.choices[0].message.content
        self._save_cache(cache_key, response_text)
        return response_text

class AzureChatOpenAI(ChatOpenAI):
    def __init__(
        self,
        model_string="gpt-35-turbo",
        system_prompt=ChatOpenAI.DEFAULT_SYSTEM_PROMPT,
        **kwargs):
        """
        Initializes an interface for interacting with Azure's OpenAI models.

        This class extends the ChatOpenAI class to use Azure's OpenAI API instead of OpenAI's API. It sets up the necessary client with the appropriate API version, API key, and endpoint from environment variables.

        :param model_string: The model identifier for Azure OpenAI. Defaults to 'gpt-3.5-turbo'.
        :param system_prompt: The default system prompt to use when generating responses. Defaults to ChatOpenAI's default system prompt.
        :param kwargs: Additional keyword arguments to pass to the ChatOpenAI constructor.

        Environment variables:
        - AZURE_OPENAI_API_KEY: The API key for authenticating with Azure OpenAI.
        - AZURE_OPENAI_API_BASE: The base URL for the Azure OpenAI API.
        - AZURE_OPENAI_API_VERSION: The API version to use. Defaults to '2023-07-01-preview' if not set.

        Raises:
            ValueError: If the AZURE_OPENAI_API_KEY environment variable is not set.
        """
        root = platformdirs.user_cache_dir("textgrad")
        cache_path = os.path.join(root, f"cache_azure_{model_string}.db")  # Changed cache path to differentiate from OpenAI cache

        super().__init__(cache_path=cache_path, system_prompt=system_prompt, **kwargs)

        self.system_prompt = system_prompt
        api_version = os.getenv("AZURE_OPENAI_API_VERSION", "2023-07-01-preview")
        if os.getenv("AZURE_OPENAI_API_KEY") is None:
            raise ValueError("Please set the AZURE_OPENAI_API_KEY, AZURE_OPENAI_API_BASE, and AZURE_OPENAI_API_VERSION environment variables if you'd like to use Azure OpenAI models.")
        
        # self.client = AzureOpenAI(
        #     api_version=api_version,
        #     api_key=os.getenv("AZURE_OPENAI_API_KEY"),
        #     azure_endpoint=os.getenv("AZURE_OPENAI_API_BASE"),
        #     azure_deployment=model_string,
        # )
        self.model_string = model_string



import os
import subprocess
import sys


def main():
    script_dir = os.path.dirname(os.path.abspath(__file__))
    script_path = os.path.join(script_dir, "Web", "🏠Overview.py")
    sys.exit(subprocess.run(["streamlit", "run", script_path]).returncode)


if __name__ == "__main__":
    main()



from typing import List, Tuple

from .base_model import BaseModel

import torch
import torch.nn.functional as F
from transformers import AutoModelForCausalLM, AutoTokenizer


class Gemma2(BaseModel):
    def __init__(self, model: AutoModelForCausalLM, tokenizer: AutoTokenizer, *args, **kwargs):
        super().__init__(model, tokenizer, *args, **kwargs)
        self.device = self.model.device
    

    def post_process(self, inputs: torch.Tensor) -> str:
        response = self.tokenizer.decode(inputs[0], skip_special_tokens=True)

        return response

    
    def forward(self, inputs: dict, generation_config: dict) -> dict:
        attention_mask = torch.ones_like(inputs, dtype=torch.long, device=inputs.device)
        outputs = self.model(inputs, attention_mask=attention_mask, **generation_config)

        return outputs
    

    def generate(self, inputs: dict, generation_config: dict) -> dict:
        attention_mask = torch.ones_like(inputs, dtype=torch.long, device=inputs.device)
        outputs = self.model.generate(inputs, attention_mask=attention_mask, **generation_config)
        response = self.post_process(outputs)

        return response
    

    def get_logits(self, input: dict, generate_config: dict) -> torch.Tensor:
        outputs = self.forward(input, generate_config)
        logits = outputs.logits

        return logits
    

    def get_candidates(self, input: dict, optimize_config: dict) -> List[int]:
        generate_config = optimize_config["generate_config"]
        logits = self.get_logits(input, generate_config)[:, -1, :]

        topk = optimize_config.get("candidates_topk", 3)
        probs = F.softmax(logits, dim=-1)
        topk_probs, topk_tokens = torch.topk(probs, topk)
        candidates = topk_tokens[0].tolist()

        return candidates
    
    @torch.inference_mode()
    def filter(self, prompts: List[Tuple[str, float]]) -> List[Tuple[str, float]]:
        readable_prompts = []

        for prompt, score in prompts:
            input_text = f'{prompt}\n Is this a human readable prompt? Only respond with yes or no.'
            input_ids = self.tokenizer(input_text, return_tensors="pt").to(self.device)
            outputs = self.model.generate(**input_ids, max_new_tokens=24)
            response = self.post_process(outputs)
            if "yes" in response.lower():
                readable_prompts.append((prompt, score))

        return readable_prompts



from .engine import EngineLM, get_engine
from typing import Union

class SingletonBackwardEngine:
    """
    A singleton class representing the backward engine.

    This class ensures that only one instance of the backward engine is created and provides methods to set and get the engine."""

    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(SingletonBackwardEngine, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        if not hasattr(self, 'engine'):
            self.engine: EngineLM = None
    
    def set_engine(self, engine: EngineLM, override: bool = False):
        """
        Sets the backward engine.

        :param engine: The backward engine to set.
        :type engine: EngineLM
        :param override: Whether to override the existing engine if it is already set. Defaults to False.
        :type override: bool
        :raises Exception: If the engine is already set and override is False.
        :return: None
        """
        if ((self.engine is not None) and (not override)):
            raise Exception("Engine already set. Use override=True to override cautiously.")
        self.engine = engine

    def get_engine(self):
        """
        Returns the backward engine.

        :return: The backward engine.
        :rtype: EngineLM
        """
        return self.engine

def set_backward_engine(engine: Union[EngineLM, str], override: bool = False):
    singleton_backward_engine = SingletonBackwardEngine()
    if isinstance(engine, str):
        engine = get_engine(engine)
    singleton_backward_engine.set_engine(engine, override=override)


def validate_engine_or_get_default(engine):
    if (engine is None) and (SingletonBackwardEngine().get_engine() is None):
        raise Exception(
            "No engine provided. Either provide an engine as the argument to this call, or use `textgrad.set_backward_engine(engine)` to set the backward engine.")
    elif engine is None:
        engine = SingletonBackwardEngine().get_engine()
    if isinstance(engine, str):
        engine = get_engine(engine)
    return engine


## Operations over variables.
from typing import List, Set
from .. import logger
from ..variable import Variable
from ..engine import EngineLM
from .reduce_prompts import construct_reduce_prompt, REDUCE_MEAN_SYSTEM_PROMPT
from .function import Function, BackwardContext


def _reduce_gradients_mean(gradients: Set[Variable], backward_engine: EngineLM) -> Variable:
    """A function to reduce gradients by taking the "mean" of the gradients. 
    In this case, we use a summarization model to summarize the gradients.
    This can be helpful in batch training, where we want to average the gradients over a batch.

    :param gradients: Gradients to summarize together
    :type gradients: Set[Variable]
    :param backward_engine: The engine to summarize gradients. 
    :type backward_engine: EngineLM
    :return: The reduce(summarized) gradients
    :rtype: Variable
    """
    gradient_reduce_prompt = construct_reduce_prompt(gradients)
    reduced_gradients = backward_engine(gradient_reduce_prompt, system_prompt=REDUCE_MEAN_SYSTEM_PROMPT)
    gradient_descriptions = set([g.get_role_description() for g in gradients])
    gradient_descriptions = ", ".join(gradient_descriptions)
    reduced_gradients_variable = Variable(reduced_gradients, role_description=gradient_descriptions)
    logger.info(f"Reduced gradients", extra={"reduced_gradients": reduced_gradients_variable.value})
    # TODO: We need to add context for these gradients
    # Otherwise, .get_gradient_and_context_text() will return an empty string
    logger.info(f"Reduced gradients", extra={"reduced_gradients": reduced_gradients_variable.value})
    return reduced_gradients_variable


class Sum(Function):
    """
    Represents a sum operation on a list of variables. 
    In TextGrad, sum is simply concatenation of the values of the variables.

    :param variables: The list of variables to be summed (concatenated).
    :type variables: List[Variable]
    :return: A new variable representing the sum of the input variables.
    :rtype: Variable
    """

    def forward(self, variables: List[Variable]) -> Variable:
        """
        Performs the forward pass of the sum (concatenation) operation.

        :param variables: The list of variables to be summed.
        :type variables: List[Variable]
        :return: A new variable representing the sum of the input variables.
        :rtype: Variable
        """
        concat_values = "\n".join([v.get_value() for v in variables])
        role_descriptions = set([v.get_role_description() for v in variables])
        role_descriptions = ", ".join(role_descriptions)
        
        total = Variable(
            value=concat_values,
            predecessors=variables,
            role_description=f"a combination of the following: {role_descriptions}",
            requires_grad=any([v.requires_grad for v in variables]),
        )
        
        total.set_grad_fn(BackwardContext(backward_fn=self.backward,
                                          summation=total))
        
        return total

        
    def backward(self, summation: Variable, backward_engine: EngineLM):
        """
        Performs the backward pass of the sum operation.
        This is simply an idempotent operation, where we pass the feedback to the predecessors variables.

        :param summation: The variable representing the sum.
        :type summation: Variable
        :param backward_engine: The backward engine used for backpropagation.
        :type backward_engine: EngineLM
        """
        children_variables = summation.predecessors
        summation_gradients = summation.get_gradient_text()
        for variable in children_variables: 
            if summation_gradients == "":
                variable_gradient_value = ""
            else:
                variable_gradient_value = f"Here is the combined feedback we got for this specific {variable.get_role_description()} and other variables: {summation_gradients}."
                
            logger.info(f"Idempotent backward", extra={"v_gradient_value": variable_gradient_value, 
                                                    "summation_role": summation.get_role_description()})

            var_gradients = Variable(value=variable_gradient_value, 
                                    role_description=f"feedback to {variable.get_role_description()}")
            variable.gradients.add(var_gradients)
            
            if summation._reduce_meta != []:
                var_gradients._reduce_meta.extend(summation._reduce_meta)
                variable._reduce_meta.extend(summation._reduce_meta)

            variable.gradients.add(Variable(value=variable_gradient_value, 
                                            role_description=f"feedback to {variable.get_role_description()}"))


class Aggregate(Function):
    """This function is WIP
    """
    def forward(self, variables: List[Variable]) -> Variable:
        """
        Aggregates a list of variables.
        In TextGrad, forward pass of aggregation is simply concatenation of the values of the variables.
        The backward pass performs a reduction operation on the gradients of the variables.
        This reduction is currently an LLM call to summarize the gradients.

        :param variables: The list of variables to be aggregated.
        :type variables: List[Variable]
        :return: The aggregated variable.
        :rtype: Variable
        """
        concat_values = "\n".join([v.get_value() for v in variables])
        role_descriptions = set([v.get_role_description() for v in variables])
        role_descriptions = ", ".join(role_descriptions)
        
        # We create a unique meta tag that identifies which variables are aggregated together.
        # We also need to communicate to the variables that they are part of a mean operation.
        reduce_meta = {"op": _reduce_gradients_mean, "id": id(variables)}
        
        aggregated_variable = Variable(value=concat_values, 
                                    role_description=f"a combination of the following variables: {role_descriptions}.",
                                    predecessors=variables,
                                    requires_grad=any([v.requires_grad for v in variables]))
        
        aggregated_variable.set_grad_fn(BackwardContext(backward_fn=self.backward,
                                                        aggregated_variable=aggregated_variable))
        
        aggregated_variable._reduce_meta = [reduce_meta]
        return aggregated_variable
        
    def backward(self, aggregated_variable: Variable, backward_engine: EngineLM):
        children_variable = aggregated_variable.predecessors
        for variable in children_variable:
            aggregate_gradients = aggregated_variable.get_gradient_text()
        if aggregate_gradients == "":
            variable_gradient_value = ""
        else:
            variable_gradient_value = f"Here is the combined feedback we got for this specific {variable.get_role_description()} and other variables: {aggregate_gradients}."
            
        logger.info(f"aggregation backward", extra={"v_gradient_value": variable_gradient_value, 
                                                "aggregation_role": aggregated_variable.get_role_description()})

        var_gradients = Variable(value=variable_gradient_value, 
                                role_description=f"feedback to {variable.get_role_description()}")
        variable.gradients.add(var_gradients)
        
        if aggregated_variable._reduce_meta != []:
            var_gradients._reduce_meta.extend(aggregated_variable._reduce_meta)
            variable._reduce_meta.extend(aggregated_variable._reduce_meta)



from . import logger
from .engine import EngineLM
from typing import List, Set, Dict
import httpx
from collections import defaultdict
from functools import partial
from .config import SingletonBackwardEngine
from .utils.image_utils import is_valid_url
from typing import Union


class Variable:
    def __init__(
            self,
            value: Union[str, bytes] = "",
            image_path: str = "",
            predecessors: List['Variable'] = None,
            requires_grad: bool = True,
            *,
            role_description: str):
        """The main thing. Nodes in the computation graph. Really the heart and soul of textgrad.

        :param value: The string value of this variable, defaults to "". In the future, we'll go multimodal, for sure!
        :type value: str or bytes, optional
        :param image_path: The path to the image file, defaults to "". If present we will read from disk or download the image.
        :type image_path: str, optional
        :param predecessors: predecessors of this variable in the computation graph, defaults to None. Here, for instance, if we have a prompt -> response through an LLM call, we'd call the prompt the predecessor, and the response the successor. 
        :type predecessors: List[Variable], optional
        :param requires_grad: Whether this variable requires a gradient, defaults to True. If False, we'll not compute the gradients on this variable.
        :type requires_grad: bool, optional
        :param role_description: The role of this variable. We find that this has a huge impact on the optimization performance, and being specific often helps quite a bit!
        :type role_description: str
        """

        if predecessors is None:
            predecessors = []

        _predecessor_requires_grad = [v for v in predecessors if v.requires_grad]

        if (not requires_grad) and (len(_predecessor_requires_grad) > 0):
            raise Exception("If the variable does not require grad, none of its predecessors should require grad."
                            f"In this case, following predecessors require grad: {_predecessor_requires_grad}")

        assert type(value) in [str, bytes, int], "Value must be a string, int, or image (bytes). Got: {}".format(
            type(value))
        if isinstance(value, int):
            value = str(value)
        # We'll currently let "empty variables" slide, but we'll need to handle this better in the future.
        # if value == "" and image_path == "":
        #    raise ValueError("Please provide a value or an image path for the variable")
        if value != "" and image_path != "":
            raise ValueError("Please provide either a value or an image path for the variable, not both.")

        if image_path != "":
            if is_valid_url(image_path):
                self.value = httpx.get(image_path).content
            else:
                with open(image_path, 'rb') as file:
                    self.value = file.read()
        else:
            self.value = value

        self.gradients: Set[Variable] = set()
        self.gradients_context: Dict[Variable, str] = defaultdict(lambda: None)
        self.grad_fn = None
        self.role_description = role_description
        self.predecessors = set(predecessors)
        self.requires_grad = requires_grad
        self._reduce_meta = []

        if requires_grad and (type(value) == bytes):
            raise ValueError("Gradients are not yet supported for image inputs. Please provide a string input instead.")

    def __repr__(self):
        return f"Variable(value={self.value}, role={self.get_role_description()}, grads={self.gradients})"

    def __str__(self):
        return str(self.value)

    def __add__(self, to_add):
        # For now, let's just assume variables can be passed to models
        if isinstance(to_add, Variable):
            ### Somehow handle the addition of variables
            total = Variable(
                value=self.value + to_add.value,
                # Add the predecessors of both variables
                predecessors=[self, to_add],
                # Communicate both of the roles
                role_description=f"{self.role_description} and {to_add.role_description}",
                # We should require grad if either of the variables require grad
                requires_grad=(self.requires_grad | to_add.requires_grad),
            )
            total.set_grad_fn(partial(
                _backward_idempotent,
                variables=total.predecessors,
                summation=total,
            ))
            return total
        else:
            return to_add.__add__(self)

    def set_role_description(self, role_description):
        self.role_description = role_description

    def reset_gradients(self):
        self.gradients = set()
        self.gradients_context = dict()
        self._reduce_meta = []

    def get_role_description(self) -> str:
        return self.role_description

    def get_short_value(self, n_words_offset: int = 10) -> str:
        """
        Returns a short version of the value of the variable. We sometimes use it during optimization, when we want to see the value of the variable, but don't want to see the entire value.
        This is sometimes to save tokens, sometimes to reduce repeating very long variables, such as code or solutions to hard problems.
        :param n_words_offset: The number of words to show from the beginning and the end of the value.
        :type n_words_offset: int
        """
        words = self.value.split(" ")
        if len(words) <= 2 * n_words_offset:
            return self.value
        short_value = " ".join(words[:n_words_offset]) + " (...) " + " ".join(words[-n_words_offset:])
        return short_value

    def get_value(self):
        return self.value

    def set_value(self, value):
        self.value = value

    def set_grad_fn(self, grad_fn):
        self.grad_fn = grad_fn

    def get_grad_fn(self):
        return self.grad_fn

    def get_gradient_text(self) -> str:
        """Aggregates and returns the gradients on a variable."""

        return "\n".join([g.value for g in self.gradients])

    def backward(self, engine: EngineLM = None):
        """
        Backpropagate gradients through the computation graph starting from this variable.

        :param engine: The backward engine to use for gradient computation. If not provided, the global engine will be used.
        :type engine: EngineLM, optional

        :raises Exception: If no backward engine is provided and no global engine is set.
        :raises Exception: If both an engine is provided and the global engine is set.
        """
        if ((engine is None) and (SingletonBackwardEngine().get_engine() is None)):
            raise Exception(
                "No backward engine provided. Either provide an engine as the argument to this call, or use `textgrad.set_backward_engine(engine)` to set the backward engine.")
        elif ((engine is not None) and (SingletonBackwardEngine().get_engine() is not None)):
            raise Exception("Both an engine is provided and the global engine is set. Be careful when doing this.")

        backward_engine = engine if engine else SingletonBackwardEngine().get_engine()
        """Taken from https://github.com/karpathy/micrograd/blob/master/micrograd/engine.py"""
        # topological order all the predecessors in the graph
        topo = []
        visited = set()

        def build_topo(v):
            if v not in visited:
                visited.add(v)
                for predecessor in v.predecessors:
                    build_topo(predecessor)
                topo.append(v)

        build_topo(self)

        # go one variable at a time and apply the chain rule to get its gradient
        # TODO: we should somehow ensure that we do not have cases such as the predecessors of a variable requiring a gradient, but the variable itself not requiring a gradient

        self.gradients = set()
        for v in reversed(topo):
            if v.requires_grad:
                v.gradients = _check_and_reduce_gradients(v, backward_engine)
                if v.get_grad_fn() is not None:
                    v.grad_fn(backward_engine=backward_engine)

    def generate_graph(self, print_gradients: bool = False):
        """
        Generates a computation graph starting from the variable itself.

        :param print_gradients: A boolean indicating whether to print gradients in the graph.
        :return: A visualization of the computation graph.
        """
        try:
            from graphviz import Digraph
        except ImportError:
            raise ImportError(
                "Please install graphviz to visualize the computation graphs. You can install it using `pip install graphviz`.")

        def wrap_text(text, width=40):
            """Wraps text at a given number of characters using HTML line breaks."""
            words = text.split()
            wrapped_text = ""
            line = ""
            for word in words:
                if len(line) + len(word) + 1 > width:
                    wrapped_text += line + "<br/>"
                    line = word
                else:
                    if line:
                        line += " "
                    line += word
            wrapped_text += line
            return wrapped_text

        def wrap_and_escape(text, width=40):
            return wrap_text(text.replace("<", "&lt;").replace(">", "&gt;"), width)

        topo = []
        visited = set()

        def build_topo(v):
            if v not in visited:
                visited.add(v)
                for predecessor in v.predecessors:
                    build_topo(predecessor)
                topo.append(v)

        def get_grad_fn_name(name):
            ws = name.split(" ")
            ws = [w for w in ws if "backward" in w]
            return " ".join(ws)

        build_topo(self)

        graph = Digraph(comment='Computation Graph starting from {}'.format(self.role_description))
        graph.attr(rankdir='TB')  # Set the graph direction from top to bottom
        graph.attr(ranksep='0.2')  # Adjust the spacing between ranks
        graph.attr(bgcolor='lightgrey')  # Set the background color of the graph
        graph.attr(fontsize='7.5')  # Set the font size of the graph

        for v in reversed(topo):
            # Add each node to the graph
            label_color = 'darkblue'

            node_label = (
                f"<b><font color='{label_color}'>Role: </font></b> {wrap_and_escape(v.role_description.capitalize())}"
                f"<br/><b><font color='{label_color}'>Value: </font></b> {wrap_and_escape(v.value)}"
            )

            if v.grad_fn is not None:
                node_label += f"<br/><b><font color='{label_color}'>Grad Fn: </font></b> {wrap_and_escape(get_grad_fn_name(str(v.grad_fn)))}"

            if v._reduce_meta != []:
                node_label += f"<br/><b><font color='{label_color}'>Reduce Meta: </font></b> {wrap_and_escape(str(v._reduce_meta))}"

            if print_gradients:
                node_label += f"<br/><b><font color='{label_color}'>Gradients: </font></b> {wrap_and_escape(v.get_gradient_text())}"
            # Update the graph node with modern font and better color scheme
            graph.node(
                str(id(v)),
                label=f"<{node_label}>",
                shape='rectangle',
                style='filled',
                fillcolor='lavender',
                fontsize='8',
                fontname="Arial",
                margin='0.1',
                pad='0.1',
                width='1.2',
            )
            # Add forward edges from predecessors to the parent
            for predecessor in v.predecessors:
                graph.edge(str(id(predecessor)), str(id(v)))

        return graph


def _check_and_reduce_gradients(variable: Variable, backward_engine: EngineLM) -> Set[Variable]:
    """
    Check and reduce gradients for a given variable.

    This function checks if the gradients of the variable need to be reduced based on the reduction groups
    specified in the variable's metadata. If reduction is required, it performs the reduction operation
    for each reduction group and returns the reduced gradients.

    For example, we do things like averaging the gradients using this functionality.

    :param variable: The variable for which gradients need to be checked and reduced.
    :type variable: Variable
    :param backward_engine: The backward engine used for gradient computation.
    :type backward_engine: EngineLM

    :return: The reduced gradients for the variable.
    :rtype: Set[Variable]
    """
    if variable._reduce_meta == []:
        return variable.gradients
    if variable.get_gradient_text() == "":
        return variable.gradients

    if len(variable.gradients) == 1:
        return variable.gradients

    id_to_gradient_set = defaultdict(set)
    id_to_op = {}  # Note: there must be a way to ensure that the op is the same for all the variables with the same id

    # Go through each gradient, group them by their reduction groups
    for gradient in variable.gradients:
        for reduce_item in gradient._reduce_meta:
            id_to_gradient_set[reduce_item["id"]].add(gradient)
            id_to_op[reduce_item["id"]] = reduce_item["op"]
    # For each reduction group, perform the reduction operation
    new_gradients = set()
    for group_id, gradients in id_to_gradient_set.items():
        logger.info(f"Reducing gradients for group {group_id}", extra={"gradients": gradients})
        new_gradients.add(id_to_op[group_id](gradients, backward_engine))

    return new_gradients


def _backward_idempotent(variables: List[Variable], summation: Variable, backward_engine: EngineLM):
    """
    Perform an idempotent backward pass e.g. for textgrad.sum or Variable.__add__.
    In particular, this function backpropagates the gradients of the `summation` variable to all the variables in the `variables` list.

    :param variables: The list of variables to backpropagate the gradients to.
    :type variables: List[Variable]
    :param summation: The variable representing the summation operation.
    :type summation: Variable
    :param backward_engine: The backward engine used for backpropagation.
    :type backward_engine: EngineLM

    :return: None

    :notes:
        - The idempotent backward pass is used for textgrad.sum or Variable.__add__ operations.
        - The gradients of the `summation` variable are backpropagated to all the variables in the `variables` list.
        - The feedback from each variable is stored in their respective gradients.
        - The feedback from the `summation` variable is combined and stored in the `summation_gradients` variable.
        - The feedback from each variable is later used for feedback propagation to other variables in the computation graph.
        - The `_reduce_meta` attribute of the `summation` variable is used to reduce the feedback if specified.
    """
    summation_gradients = summation.get_gradient_text()
    for variable in variables:
        if summation_gradients == "":
            variable_gradient_value = ""
        else:
            variable_gradient_value = f"Here is the combined feedback we got for this specific {variable.get_role_description()} and other variables: {summation_gradients}."

        logger.info(f"Idempotent backward", extra={"v_gradient_value": variable_gradient_value,
                                                   "summation_role": summation.get_role_description()})

        var_gradients = Variable(value=variable_gradient_value,
                                 role_description=f"feedback to {variable.get_role_description()}")
        variable.gradients.add(var_gradients)

        if summation._reduce_meta != []:
            var_gradients._reduce_meta.extend(summation._reduce_meta)
            variable._reduce_meta.extend(summation._reduce_meta)

        variable.gradients.add(Variable(value=variable_gradient_value,
                                        role_description=f"feedback to {variable.get_role_description()}"))



GLOSSARY_TEXT = """
### Glossary of tags that will be sent to you:
# - <LM_SYSTEM_PROMPT>: The system prompt for the language model.
# - <LM_INPUT>: The input to the language model.
# - <LM_OUTPUT>: The output of the language model.
# - <FEEDBACK>: The feedback to the variable.
# - <CONVERSATION>: The conversation history.
# - <FOCUS>: The focus of the optimization.
# - <ROLE>: The role description of the variable."""

### Optimize Prompts

# System prompt to TGD
OPTIMIZER_SYSTEM_PROMPT = (
    "You are part of an optimization system that improves text (i.e., variable). "
    "You will be asked to creatively and critically improve prompts, solutions to problems, code, or any other text-based variable. "
    "You will receive some feedback, and use the feedback to improve the variable. "
    "The feedback may be noisy, identify what is important and what is correct. "
    "Pay attention to the role description of the variable, and the context in which it is used. "
    "This is very important: You MUST give your response by sending the improved variable between {new_variable_start_tag} {{improved variable}} {new_variable_end_tag} tags. "
    "The text you send between the tags will directly replace the variable.\n\n"
    f"{GLOSSARY_TEXT}"
)

# TGD update instruction
TGD_PROMPT_PREFIX = (
    "Here is the role of the variable you will improve: <ROLE>{variable_desc}</ROLE>.\n\n"
    "The variable is the text within the following span: <VARIABLE> {variable_short} </VARIABLE>\n\n"
    "Here is the context and feedback we got for the variable:\n\n"
    "<CONTEXT>{variable_grad}</CONTEXT>\n\n"
    "Improve the variable ({variable_desc}) using the feedback provided in <FEEDBACK> tags.\n"
)

# If the gradients are in a multi-part container
TGD_MULTIPART_PROMPT_INIT = (
    "Here is the role of the variable you will improve: <ROLE>{variable_desc}</ROLE>.\n\n"
    "The variable is the text within the following span: <VARIABLE> {variable_short} </VARIABLE>\n\n"
    "Here is the context and feedback we got for the variable:\n\n"
)

TGD_MULTIPART_PROMPT_PREFIX = (
    "Improve the variable ({variable_desc}) using the feedback provided in <FEEDBACK> tags.\n"
)

TGD_PROMPT_SUFFIX  = (
    "Send the improved variable "
    "in the following format:\n\n{new_variable_start_tag}{{the improved variable}}{new_variable_end_tag}\n\n"
    "Send ONLY the improved variable between the <IMPROVED_VARIABLE> tags, and nothing else."
)

MOMENTUM_PROMPT_ADDITION = (
    "Here are the past iterations of this variable:\n\n"
    "<PAST_ITERATIONS>{past_values}</PAST_ITERATIONS>\n\n"
    "Similar feedbacks across different steps suggests that the modifications to the variable are insufficient." 
    "If this is the case, please make more significant changes to the variable.\n\n"
)

CONSTRAINT_PROMPT_ADDITION = (
    "You must follow the following constraints:\n\n"
    "<CONSTRAINTS>{constraint_text}</CONSTRAINTS>\n\n"
)

IN_CONTEXT_EXAMPLE_PROMPT_ADDITION = (
    "You must base on the following examples when modifying the {variable_desc}:\n\n"
    "<EXAMPLES>{in_context_examples}</EXAMPLES>\n\n"
)

def construct_tgd_prompt(do_momentum: bool = False,
                         do_constrained: bool = False,
                         do_in_context_examples: bool = False,
                         **optimizer_kwargs):
    """
    Construct the textual gradient descent prompt.

    :param do_momentum: Whether to include momentum in the prompt.
    :type do_momentum: bool, optional
    :param do_constrained: Whether to include constraints in the prompt.
    :type do_constrained: bool, optional
    :param do_in_context_examples: Whether to include in-context examples in the prompt.
    :type do_in_context_examples: bool, optional
    :param optimizer_kwargs: Additional keyword arguments for formatting the prompt. These will be things like the variable description, gradient, past values, constraints, and in-context examples.
    :return: The TGD update prompt.
    :rtype: str
    """

    if isinstance(optimizer_kwargs["variable_grad"], str):
        multipart=False
        prompt = TGD_PROMPT_PREFIX.format(**optimizer_kwargs)
        
    else:
        gradient_context = optimizer_kwargs["variable_grad"]
        gradient_context = [TGD_MULTIPART_PROMPT_INIT.format(**optimizer_kwargs)] + gradient_context
        multipart=True
        prompt = TGD_MULTIPART_PROMPT_PREFIX.format(**optimizer_kwargs)
           
    if do_momentum:
        prompt += MOMENTUM_PROMPT_ADDITION.format(**optimizer_kwargs)

    if do_constrained:
        prompt += CONSTRAINT_PROMPT_ADDITION.format(**optimizer_kwargs)

    if do_in_context_examples:
        prompt += IN_CONTEXT_EXAMPLE_PROMPT_ADDITION.format(**optimizer_kwargs)

    prompt += TGD_PROMPT_SUFFIX.format(**optimizer_kwargs)

    if not multipart:
        return prompt
    
    else:
        return gradient_context + [prompt]

# This is how we save gradients to the variable.
GRADIENT_TEMPLATE = (
    "Here is a conversation:\n\n<CONVERSATION>{context}</CONVERSATION>\n\n"
    "This conversation is potentially part of a larger system. The output is used as {response_desc}\n\n"
    "Here is the feedback we got for {variable_desc} in the conversation:\n\n<FEEDBACK>{feedback}</FEEDBACK>\n\n"
)
GRADIENT_MULTIPART_TEMPLATE = (
    "Above is a conversation with a language model.\n"
    "This conversation is potentially part of a larger system. The output is used as {response_desc}\n\n"
    "Here is the feedback we got for {variable_desc} in the conversation:\n\n<FEEDBACK>{feedback}</FEEDBACK>\n\n"
)



import os
import json

class Node:
    def __init__(self, timestamp, id, prompt, history=None, reasoning=None,
                 parent=None, optim_batch=None, 
                 train_results=None, train_score=None, 
                 dev_results=None, dev_score=None):
        
        self.timestamp = timestamp
        self.id = id # a string
        self.prompt = prompt
        self.history = history

        self.parent = parent # id of the parent node
        self.optim_batch = optim_batch # indices of training examples that are used to update parent -> this node
        
        self.scores = {} # split_name: score on the split
        self.results = {} # split_name: dataframe of the split

        self.n_child = 0
        self.reasoning = reasoning

    def __str__(self):
        s = "Timestamp: {}".format(self.timestamp)
        s += "\nID: {}".format(self.id)
        s += "\nPrompt: {}".format(self.prompt[:200])
        if len(self.prompt) > 200:
            s += "\n[Displaying the first 200 characters of the prompt ...]"
        return s
    
    def to_dict(self):
        d = {
            "timestamp": self.timestamp,
            "id": self.id,
            "parent": self.parent,
            "scores": self.scores,
            "prompt": self.prompt,
            "history": self.history,
            "optim_batch": self.optim_batch,
            "reasoning": self.reasoning,
        }
            
        return d

    def save(self, out_dir):
        node_dir = os.path.join(out_dir, self.id)
        os.makedirs(node_dir, exist_ok=True)
        
        # full node
        node_file = os.path.join(node_dir, "node.json")
        d = self.to_dict()
        with open(node_file, "w") as f:
            json.dump(d, f, indent=4)
            
        # prompt
        prompt_file = os.path.join(node_dir, "prompt.md")
        with open(prompt_file, "w") as f:
            f.write(self.prompt)
            
        # results (if any)
        for key, value in self.results.items():
            result_file = os.path.join(node_dir, "{}_result.csv".format(key))
            value.to_csv(result_file)
            
    def register_results(self, results, score, split_name):
        self.scores[split_name] = score
        self.results[split_name] = results


from .utils import get_llm

class AbstractModel:
    def __init__(self, args, logger):
        self.args = args
        self.logger = logger

    def load_prompt(self):
        raise NotImplementedError
    
    def run(self, data):
        raise NotImplementedError


from .mmlu import MMLU, MMLUInstanceDataset
from .base import Dataset, DataLoader
from .leetcode import LeetCodeHardEval

from typing import Tuple, Callable
from ..variable import Variable
from ..engine import EngineLM

AVAILABLE_DATASETS = [
    "BBH_object_counting",
    "BBH_word_sorting",
    "GSM8K_DSPy",
]

AVAILABLE_INSTANCE_DATASETS = [
    "MMLU_machine_learning",
    "MMLU_college_physics",
    "GPQA_diamond"
    "LeetCodeHardEval"
]

def load_task(p_init, task_name: str, evaluation_api: EngineLM, *args, **kwargs) -> Tuple[Dataset, Dataset, Callable]:
    """
    Args:
        task_name: the name of the task to evaluate
        evaluation_api: the engine to use for evaluation, if needed
    """
    if "object_counting" in task_name:
        from ..loss import MultiFieldTokenParsedEvaluation
        from .big_bench_hard import BigBenchHard, string_based_equality_fn
        from ..autograd.string_based_ops import StringBasedFunction
        task_name = task_name[4:]
        train_set = BigBenchHard(p_init, task_name, split="train", *args, **kwargs)
        val_set = BigBenchHard(p_init, task_name, split="val", *args, **kwargs)
        test_set = BigBenchHard(p_init, task_name, split="test", *args, **kwargs)
        role_descriptions = [
            "Question for the task",
            "Ground truth answer",
            "Reasoning and prediction from the language model"
        ]
        fn_purpose = "The runtime of string-based function that checks if the prediction is correct."
        eval_fn = StringBasedFunction(string_based_equality_fn, function_purpose=fn_purpose)
        return train_set, val_set, test_set, eval_fn, train_set.csv_data_path
    
    elif "BBH" in task_name:
        from textgrad.loss import MultiFieldTokenParsedEvaluation
        from .big_bench_hard import BigBenchHard
        task_name = task_name[4:]
        train_set = BigBenchHard(p_init, task_name, split="train", *args, **kwargs)
        val_set = BigBenchHard(p_init, task_name, split="val", *args, **kwargs)
        test_set = BigBenchHard(p_init, task_name, split="test", *args, **kwargs)
        role_descriptions = [
            "Question for the task",
            "Ground truth answer",
            "Reasoning and prediction from the language model"
        ]

        # invalid for smaller and less powerful models
        evaluation_instruction = "Below is a question from a question-answering task, the ground truth answer, and reasoning with the final prediction. Is the final prediction correct, i.e. the same as the ground truth answer? Say only 1 (yes) or 0 (no). Return your response within <ACCURACY> </ACCURACY> tags. e.g.<ACCURACY> 0 </ACCURACY> or <ACCURACY> 1 </ACCURACY>"
        eval_instruction = Variable(evaluation_instruction, requires_grad=False, role_description="evaluation instruction for the task")
        eval_fn = MultiFieldTokenParsedEvaluation(
            eval_instruction,
            engine=evaluation_api,
            role_descriptions=role_descriptions,
            parse_tags=["<ACCURACY>", "</ACCURACY>"]
        )
        
        return train_set, val_set, test_set, eval_fn, train_set.csv_data_path
    
    elif task_name == "GSM8K_DSPy":
        from textgrad.tasks.gsm8k import GSM8K_DSPy
        from .big_bench_hard import string_based_equality_fn
        from textgrad.autograd.string_based_ops import StringBasedFunction
        evaluation_instruction = "Below is a prediction we got for a question answering task, and the correct final answer. Is the final answer correct? Say only 1 (yes) or 0 (no). Return 1 if and only if the final answer is correct. Return your response within <ACCURACY> </ACCURACY> tags. e.g.<ACCURACY> 0 </ACCURACY> or <ACCURACY> 1 </ACCURACY>"
        system_prompt = Variable("You are a language model that evaluates the accuracy of a prediction for a mathematical question answering task. Only call a prediction accurate if it is the same as the ground truth answer.", requires_grad=False, role_description="system prompt for the evaluation")
        # Should we do train/test like this?
        train_set = GSM8K_DSPy(split="train", *args, **kwargs)
        val_set = GSM8K_DSPy(split="val", *args, **kwargs)
        test_set = GSM8K_DSPy(split="test", *args, **kwargs)
        role_descriptions = [
            "Question for the task",
            "Ground truth answer",
            "Prediction from the language model"
        ]
        fn_purpose = "The runtime of string-based function that checks if the prediction is correct."
        eval_fn = StringBasedFunction(string_based_equality_fn, function_purpose=fn_purpose)
        return train_set, val_set, test_set, eval_fn
    
    else:
        raise ValueError(f"Task {task_name} not found.")


def load_instance_task(task_name: str, evaluation_api: EngineLM, *args, **kwargs):
    if "MMLU_" in task_name:
        subset = task_name[5:]
        test_set = MMLUInstanceDataset(evaluation_api=evaluation_api, subset=subset, split="test", *args, **kwargs)
        return test_set
    elif "GPQA" in task_name:
        from .gpqa import GPQAInstanceDataset
        test_set = GPQAInstanceDataset(evaluation_api=evaluation_api, subset=task_name.lower(), *args, **kwargs)
        return test_set
    elif task_name in ["LeetCodeHardEval"]:
        dataset = LeetCodeHardEval()
        return dataset
    else:
        raise ValueError(f"Instance task {task_name} not found.")


import argparse
import os
import string
import time
from typing import List, Tuple

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
PARENT_DIR = os.path.dirname(CURRENT_DIR)


def clean_string(prompts: List[Tuple[str, float]]) -> List[Tuple[str, float]]:
    cleaned_prompts = []

    for prompt, score in prompts:
        prompt = prompt.strip("'")
        prompt = prompt.rstrip('.')
        prompt = prompt.translate(str.maketrans('', '', string.punctuation))
        cleaned_prompts.append((prompt, score))
    
    cleaned_prompts = list(set(cleaned_prompts))
    cleaned_prompts.sort(key=lambda x: x[1])

    return cleaned_prompts


def ape_apo_pe2_args(trainer: str) -> argparse.Namespace:
    args = argparse.Namespace()
    args.do_train = True
    args.do_validate = True
    args.do_test = True
    args.trainer = trainer
    args.backtrack = True
    args.resume = True
    args.meta_prompts_dir = f"{PARENT_DIR}/core/pe2/meta_prompts/{trainer}"
    args.model = "zeroshotcot"
    args.task = "bbh"
    args.subtask = "boolean_expressions"
    args.data_dir = f"{PARENT_DIR}/core/pe2/data/bbh/boolean_expressions"
    args.output_dir = f"{PARENT_DIR}/core/pe2/output/boolean_expressions/{trainer}_{time.strftime('%Y%m%d_%H%M%S')}"
    args.task_model = "openai_gpt35_turbo_instruct"
    args.optim_model = "openai_gpt4_turbo"
    args.n_beam = 2
    args.n_expand = 2
    args.train_steps = 2
    args.init_temperature = 0.7
    args.init_method = "file"
    args.init_prompt_file = "prompt.md"
    args.init_n_demo = None
    args.init_n_prompts = 1
    args.prompt_max_tokens = 50
    args.batch_size = 2
    args.step_size = 10
    args.optim_use_gradient = False
    args.optim_use_momentum = False
    args.optim_use_step_size = False
    args.optim_use_instruction = False
    args.optim_use_demonstrations = False
    args.optim_use_optim_tutorial = False
    args.batching = "hard"
    args.bandit = "all"
    args.seed = 42
    args.debug = False

    return args


def textgrad_args() -> argparse.Namespace:
    args = argparse.Namespace()
    args.task = "BBH_object_counting"
    args.evaluation_engine = "meta-llama/Meta-Llama-3-8B-Instruct"
    args.test_engine = "meta-llama/Meta-Llama-3-8B-Instruct"
    args.batch_size = 2
    args.max_epochs = 3
    args.seed = 42
    args.run_validation = True
    args.do_not_run_larger_model = False
    args.num_threads = 1

    return args



import random
import guidance

def deduplicate(nodes, prev_nodes=None):
    # deduplicate prompts (there is a slight chance that two prompts are the same...)
    unique_keys = set()
    deduplicated_list = []
    if prev_nodes: # from past time stamps
        for node in prev_nodes:
            unique_keys.add(node.prompt)
    for node in nodes:
        if node.prompt not in unique_keys:
            unique_keys.add(node.prompt)
            deduplicated_list.append(node)
    return deduplicated_list 

def clean_string(s):
    # issues with substrate llm api
    return s.replace("<|im_start|>", "").replace("<|im_end|>", "").strip('"').strip()

def pack_demo_string(batch):
    d = ""
    for i, row in batch.iterrows():
        if isinstance(row["label"], list):
            label = random.choice(row["label"])
        else:
            label = row["label"]
        d += """Input: {}\nOutput: {}\n\n""".format(
            row["input"], label)
    return d

def _load_prompt(datafile: str):
    with open(datafile, "r") as f:
        prompt = f.read()
    return prompt

def get_llm(llm_name):
    assert llm_name in ["openai_gpt35", "openai_gpt4", "openai_gpt4_turbo", "openai_gpt4o", "openai_gpt4o_mini"]
    
    if llm_name == "openai_gpt35":
        llm = guidance.llms.OpenAI("gpt-3.5-turbo", max_retries=100, max_calls_per_min=10)
    elif llm_name == "openai_gpt4": # we were using gpt-4-0613 for the experiments in the paper
        llm = guidance.llms.OpenAI("gpt-4", max_retries=100, max_calls_per_min=1)
    elif llm_name == "openai_gpt4_turbo": # we were using gpt-4-0125-preview for the experiments in the paper
        llm = guidance.llms.OpenAI("gpt-4-0125-preview", max_retries=100, max_calls_per_min=3, chat_mode=True)
    elif llm_name == "openai_gpt4o": 
        llm = guidance.llms.OpenAI("gpt-4o", max_retries=100, max_calls_per_min=3, chat_mode=True)    
    elif llm_name == "openai_gpt4o_mini": 
        llm = guidance.llms.OpenAI("gpt-4o-mini", max_retries=100, max_calls_per_min=10, chat_mode=True)    
    return llm



import platformdirs

from ..variable import Variable
from ..loss import MultiChoiceTestTime, MultiFieldTokenParsedEvaluation
from .base import Dataset

import re

def eval_string_based(response_text, correct_answer):
    ANSWER_PATTERN_MULTICHOICE = r"(?i)Answer\s*:\s*([A-D])"
    
    match = re.search(ANSWER_PATTERN_MULTICHOICE, response_text)
    extracted_answer = match.group(1) if match else None
    score = 1.0 if extracted_answer == correct_answer else 0.0
    return score

# Below template is from https://github.com/openai/simple-evals/blob/main/common.py#L12
QUERY_TEMPLATE_MULTICHOICE = """
Answer the following multiple choice question. The last line of your response should be of the following format: 'Answer: $LETTER' (without quotes) where LETTER is one of ABCD. Think step by step before answering.

{Question}

A) {A}
B) {B}
C) {C}
D) {D}
""".strip()

class MMLU(Dataset):
    def __init__(self, subset:str, root: str=None, split: str="train", *args, **kwargs):
        """
        MMLU dataset from HF."""
        from datasets import load_dataset
        if root is None:
            root = platformdirs.user_cache_dir("textgrad")
            
        self.root = root
        self.subset = subset
        assert split in ["train", "validation", "test"]
        self.data = load_dataset("cais/mmlu", subset, cache_dir=root, split=split if split != "train" else "dev")
        self.split = split
        self._task_description = 'You will answer multiple-choice questions. Think step by step.'
            
    def __getitem__(self, index):
        row = self.data[index]
        question = row["question"]
        choices = row["choices"]
        # Choices will be a. Choice 1 b. Choice 2 ... etc
        choices_str = "\n".join([f"{chr(65+i)}. {choice}" for i, choice in enumerate(choices)])
        answer = chr(65+row["answer"])
        question_prompt = f"Question: {question}\nChoices:\n{choices_str}"
        return question_prompt, answer

    def __len__(self):
        return len(self.data)

    def get_default_task_instruction(self):
        return "Given a multiple choice question, the goal is to select the correct final answer from the choices."


class MMLUInstanceDataset(MMLU):
    def __init__(self, evaluation_api, subset:str, root: str=None, split: str="train", max_samples=-1):
        super().__init__(subset, root, split, max_samples)
        self.evaluation_api = evaluation_api

        
    def _get_instance_test_time_objective(self, question: str):
        evaluation_instruction = "Below is a multi-choice question and an answer. You are an expert scientist. Your job is to investigate the answer. Critically go through reasoning steps, consider your knowledge, and see if the answer is correct or if there are any critical mistakes."
        eval_fn = MultiChoiceTestTime(evaluation_instruction, engine=self.evaluation_api)
        def test_time_objective(instance: Variable):
            return eval_fn(question, instance)
        return test_time_objective
        

    def _legacy_get_instance_eval_fn(self, question_prompt: str, answer: str):
        role_descriptions = [
            "Question for the task",
            "Correct answer",
            "Prediction from the language model"
        ]
        eval_system_prompt = Variable("You are a language model that evaluates the accuracy of a prediction for a mathematical question answering task. Only call a prediction accurate if it is the same as the ground truth answer.", requires_grad=False, role_description="system prompt for the evaluation")

        evaluation_instruction = "Below is a question from a question-answering task, the ground truth answer, and a prediction. Is the final prediction correct, i.e. the same as the ground truth answer? Say only 1 (yes) or 0 (no). Return your response within <ACCURACY> </ACCURACY> tags. e.g.<ACCURACY> 0 </ACCURACY> or <ACCURACY> 1 </ACCURACY>"
        eval_instruction = Variable(evaluation_instruction, requires_grad=False, role_description="evaluation instruction for the task")
        eval_fn = MultiFieldTokenParsedEvaluation(
            eval_instruction,
            engine=self.evaluation_api,
            role_descriptions=role_descriptions,
            parse_tags=["<ACCURACY>", "</ACCURACY>"],
            system_prompt=eval_system_prompt
        )
        
        answer_var = Variable(answer, requires_grad=False, role_description="Correct answer")
        question_var = Variable(question_prompt, requires_grad=False, role_description="Question for the task")
        def instance_eval_fn(instance):
            eval_output = eval_fn([question_var, answer_var, instance])
            return eval_fn.parse_output(eval_output)
        return instance_eval_fn

    def _get_instance_eval_fn(self, question_prompt: str, answer: str):
        eval_string_based_fn = lambda response: eval_string_based(response.value, answer)
        return eval_string_based_fn
    
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, index):
        row = self.data[index]
        question = row["question"]
        choices = row["choices"]
        choices_dict = dict(
                A=choices[0], B=choices[1], C=choices[2], D=choices[3], Question=question
            )
        question_prompt = QUERY_TEMPLATE_MULTICHOICE.format(**choices_dict)

        # Choices will be a. Choice 1 b. Choice 2 ... etc
        answer = chr(65+row["answer"])
        # TODO: Make the two-way comparison class abstract enough.
        # TODO: How do we determine the role of the instances? We should be more consistent
        return question_prompt, answer, self._get_instance_test_time_objective(question_prompt), self._get_instance_eval_fn(question_prompt, answer)

    def get_default_task_instruction(self):
        return "Given a multiple choice question, the goal is to select the correct final answer from the choices."



from openai import OpenAI

## In another console, run the following,
## python -u -m vllm.entrypoints.openai.api_server --host 0.0.0.0 --port 8001 --model mistralai/Mistral-7B-Instruct-v0.2 --gpu_memory_utilization 0.8

openai_api_key = "EMPTY"
openai_api_base = "http://localhost:8001/v1"
client = OpenAI(
    api_key=openai_api_key,
    base_url=openai_api_base,
)

completion = client.completions.create(model="mistralai/Mistral-7B-Instruct-v0.2",
                                      prompt="San Francisco is a")
print("Completion result:", completion)

chat_response = client.chat.completions.create(
    model="mistralai/Mistral-7B-Instruct-v0.2",
    messages=[
        {"role": "user", "content": "Tell me a joke."},
    ]
)
print("Chat response:", chat_response)

## Expected Output
### Completion result: Completion(id='cmpl-adfdeec90cc8473eb6c29c10e2ada2cb', choices=[CompletionChoice(finish_reason='length', index=0, logprobs=None, text=' foodie paradise, with dozens of exciting, new restaurants opening each year.', stop_reason=None)], created=1722372949, model='mistralai/Mistral-7B-Instruct-v0.2', object='text_completion', system_fingerprint=None, usage=CompletionUsage(completion_tokens=16, prompt_tokens=5, total_tokens=21))
### Chat response: ChatCompletion(id='chat-eb65f598e5d64444b312bcadb82e1460', choices=[Choice(finish_reason='stop', index=0, logprobs=None, message=ChatCompletionMessage(content=" Why don't scientists trust atoms?\n\nBecause they make up everything!", role='assistant', function_call=None, tool_calls=[]), stop_reason=None)], created=1722372949, model='mistralai/Mistral-7B-Instruct-v0.2', object='chat.completion', system_fingerprint=None, usage=CompletionUsage(completion_tokens=17, prompt_tokens=13, total_tokens=30))  


import sys
sys.path.append("../src")
import json

from greaterprompt.optimizer import TextGradOptimizer, GreaterDataloader

import torch
import streamlit as st
from torch.nn import functional as F


CUDA_DEVICES = torch.cuda.device_count()

with st.sidebar:
    settings = st.markdown("<h3>🛠️ Advanced Settings</h3>", unsafe_allow_html=True)
    device = st.selectbox("Device", ["cpu", *[f"cuda:{i}" for i in range(CUDA_DEVICES)]], index=1)

st.markdown("<h1 style='text-align: center; white-space: nowrap;'>🤩 Optimize with TextGradOptimizer</h1>", unsafe_allow_html=True)
evaluation_engine = st.text_input("Evaluation Engine", key="Evaluation Engine", value='meta-llama/Meta-Llama-3-8B-Instruct')
test_engine = st.text_input("Test Engine", key="Test Engine", value='meta-llama/Meta-Llama-3-8B-Instruct')
uploaded_file = st.file_uploader("Upload a jsonl input file", type=("jsonl"))

p_init = st.text_input("P Initial", value="")

col1, col2, col3 = st.columns([1, 2, 1])
with col2:
    start_button = st.button("⚡ Start Optimization", type="primary", use_container_width=True)

if not uploaded_file and start_button:
    st.info("Please upload a jsonl file to start")

if p_init.strip() == "" and start_button:
    st.info("Please enter a valid P Initial to start")

if uploaded_file and p_init.strip() and start_button:
    content = uploaded_file.getvalue().decode("utf-8")
    lines = content.strip().split('\n')

    inputs = []
    
    for line in lines:
        if line.strip():
            try:
                inputs.append(json.loads(line))
            except json.JSONDecodeError:
                st.error(f"Invalid JSON in line: {line}")

    with st.status("⚡ Optimizing...", expanded=True) as status:
        textgrad_optimizer = TextGradOptimizer({
            "evaluation_engine": evaluation_engine,
            "test_engine": test_engine,
            "device": device
        })
        dataset = GreaterDataloader(custom_inputs=inputs)

        output = textgrad_optimizer.optimize(p_init, dataset)

        status.update(label="😄 Optimization complete!", state="complete")
        st.markdown("<h5 style='text-align: left; white-space: nowrap;'>🎯 Optimization Result:</h5>", unsafe_allow_html=True)
        st.markdown(f"<p>{output}</p>", unsafe_allow_html=True)

        if output:
            st.download_button(
                label="📥 Download Optimized Result",
                data=json.dumps(output, indent=2, ensure_ascii=False),
                file_name="optimized_prompt.json",
                mime="application/json"
            )



import torch
from transformers import (AutoModelForCausalLM, AutoTokenizer, GPT2LMHeadModel,
                          GPTJForCausalLM, GPTNeoXForCausalLM,
                          LlamaForCausalLM)

from fastchat.model import get_conversation_template
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP
import re


# TODO adjust the parameter of path
PATH = '/scratch2/share/model_files/huggingface'


class ModelClass(object):
    def __init__(self, *, model_path, is_distributed=False, use_path=True, system_prompt="", device):
        self.is_distributed = is_distributed


        self.system_prompt = system_prompt

        if is_distributed:
            dist.init_process_group(backend='nccl')
            self.rank = dist.get_rank()
            self.device = torch.device(f'cuda:{self.rank}')
            torch.cuda.set_device(self.device)
            if use_path:

                self.model = AutoModelForCausalLM.from_pretrained(
                    model_path, torch_dtype=torch.float16,
                    trust_remote_code=True, device_map=device)
            else:
                self.model = AutoModelForCausalLM.from_pretrained(
                    model_path, torch_dtype=torch.float16,
                    trust_remote_code=True, device_map=device)

            self.model_name_or_path = self.model.name_or_path

            self.model = DDP(self.model, device_ids=[self.rank])

        else:

            if use_path:

                self.model = AutoModelForCausalLM.from_pretrained(model_path, torch_dtype=torch.float16,
                                                              trust_remote_code=True, device_map=device,
                                                              cache_dir=PATH)
            else:
                self.model = AutoModelForCausalLM.from_pretrained(model_path, torch_dtype=torch.float16,trust_remote_code=True, device_map=device)

            self.model_name_or_path = self.model.name_or_path

        # self.model = AutoModelForCausalLM.from_pretrained(model_path, trust_remote_code=True, device_map="auto",
        #                                                   cache_dir=PATH)
        #
        # self.model = self.model.bfloat16()
        self.model.eval()
        if 'llama' in model_path:
            self.tokenizer = AutoTokenizer.from_pretrained(model_path, cache_dir=PATH, add_bos_token=False)
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(model_path, cache_dir = PATH)
        #self.tokenizer.pad_token = "[PAD]"

        if 'Llama-2' in model_path:
            self.tokenizer.pad_token = self.tokenizer.eos_token
            self.tokenizer.pad_id = -1

        elif 'Llama-3' in model_path:
            self.tokenizer.pad_token = self.tokenizer.eos_token
            self.tokenizer.pad_token_id = self.tokenizer.eos_token_id

        elif 'gemma-2' in model_path:
            pass

        # forced model loading across all gpus
        # so that Llama2-70B running is possible

        print("model loaded successfully")
        self.template = get_conversation_template(model_path)

    def get_answer(self, prompt, system_prompt=None, output_only=True, length = 1000, temperature=1.0, top_p=0):
        if system_prompt:
            self.system_prompt = system_prompt
        if type(prompt) == type("string"):
            prompt = [prompt]
        length = 1024
        #print(len(prompt))

        prompts_list = []

        if 'Llama-2' in self.model_name_or_path:

            if self.system_prompt != "":
                self.system_prompt = f"<SYS>>\n{self.system_prompt}\n<</SYS>>\n\n"
            for k in prompt:
                self.template.messages = []
                self.template.append_message(self.template.roles[0], f"{k}")
                self.template.append_message(self.template.roles[1], "")

                temp = self.template.get_prompt()

                pattern = """<<SYS>>\nYou are a helpful, respectful and honest assistant. Always answer as helpfully as possible, while being safe. Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content. Please ensure that your responses are socially unbiased and positive in nature.\n\nIf a question does not make any sense, or is not factually coherent, explain why instead of answering something not correct. If you don't know the answer to a question, please don't share false information.\n<</SYS>>\n\n"""



                temp = re.sub(pattern, self.system_prompt, temp)
                prompts_list.append(temp)
                #
                #
                # prompts_list.append(self.template.get_prompt())

            # Distribute data across all gpus
            if self.is_distributed:
                total_prompts = len(prompts_list)
                prompts_per_gpu = total_prompts // dist.get_world_size()
                start_idx = self.rank * prompts_per_gpu
                end_idx = start_idx + prompts_per_gpu if self.rank != dist.get_world_size() - 1 else total_prompts
                local_prompts = prompts_list[start_idx:end_idx]
                encoding = self.tokenizer.batch_encode_plus(local_prompts, padding=True, return_tensors="pt").to(
                    self.device)

            else:
                encoding = self.tokenizer.batch_encode_plus(prompts_list, padding=True, return_tensors="pt").to(self.model.device)
            #out = self.model.generate(encoding.input_ids, attention_mask=encoding.attention_mask, max_new_tokens = length, temperature=temperature)

            # debugged
            if encoding.input_ids.shape[1] + 50 > length:
                length = encoding.input_ids.shape[1] + 50

            out = self.model.generate(encoding.input_ids, attention_mask=encoding.attention_mask, do_sample=False)

            #out = out[out.find(self.template.roles[1]) + len(self.template.roles[1])]
            outs = self.tokenizer.batch_decode(out.tolist(), skip_special_tokens=True)

            #return out[out.find(self.template.roles[1]) + len(self.template.roles[1]):]
            return [out[out.find(self.template.roles[1]) + len(self.template.roles[1]):] for out in outs]

        elif 'Llama-3' in self.model_name_or_path:
            if self.system_prompt != "":
                self.system_prompt = f"<|start_header_id|>system<|end_header_id|>\n\n{self.system_prompt}<|eot_id|>\n"
            for k in prompt:
                user_message = k
                temp = f"{self.system_prompt}<|start_header_id|>user<|end_header_id|>\n\n{user_message}<|eot_id|><|start_header_id|>assistant<|end_header_id|>\n\n"
                prompts_list.append(temp)

            if self.is_distributed:
                total_prompts = len(prompts_list)
                prompts_per_gpu = total_prompts // dist.get_world_size()
                start_idx = self.rank * prompts_per_gpu
                end_idx = start_idx + prompts_per_gpu if self.rank != dist.get_world_size() - 1 else total_prompts
                local_prompts = prompts_list[start_idx:end_idx]
                encoding = self.tokenizer.batch_encode_plus(local_prompts, padding=True, return_tensors="pt").to(
                    self.device)

            else:
                encoding = self.tokenizer.batch_encode_plus(prompts_list, padding=True, return_tensors="pt").to(
                    self.model.device)

            #encoding = self.tokenizer.batch_encode_plus(prompts_list, return_tensors="pt", padding=True).to(self.model.device)
            #out = self.model.generate(encoding.input_ids, attention_mask=encoding.attention_mask, max_new_tokens=length, temperature=temperature)
            out = self.model.generate(encoding.input_ids, attention_mask=encoding.attention_mask, max_new_tokens=length,
                                     do_sample=False)
            outs = self.tokenizer.batch_decode(out.tolist(), skip_special_tokens=True)

            #return [out[len(self.tokenizer.decode(enc)):] for out, enc in zip(outs, encoding['input_ids'])]
            return [out.split("assistant")[-1].strip() for out in outs]

        elif 'gemma-2' in self.model_name_or_path:
            if self.system_prompt != "":
                self.system_prompt = f"{self.system_prompt} "
            for k in prompt:
                #user_message = f"<|startoftext|>user\n{k}\n"
                user_message = f"{k}"
                temp = f"<start_of_turn>user\n{self.system_prompt}{user_message}<end_of_turn>\n<start_of_turn>model\n"
                prompts_list.append(temp)

            encoding = self.tokenizer.batch_encode_plus(prompts_list, padding=True, return_tensors="pt").to(self.model.device)

            out = self.model.generate(encoding.input_ids, attention_mask=encoding.attention_mask, max_new_tokens=length, do_sample=False)
            outs = self.tokenizer.batch_decode(out.tolist(), skip_special_tokens=True)
            return [out.split("model\n")[-1].strip() for out in outs]

        elif 'gemma' in self.model_name_or_path:
            for k in prompt:
                #prompt_t = f"""<start_of_turn>user\n{k}<end_of_turn>\n<start_of_turn>model"""
                prompt_t = f"{k}"
                prompts_list.append(prompt_t)

            if self.is_distributed:
                total_prompts = len(prompts_list)
                prompts_per_gpu = total_prompts // dist.get_world_size()
                start_idx = self.rank * prompts_per_gpu
                end_idx = start_idx + prompts_per_gpu if self.rank != dist.get_world_size() - 1 else total_prompts
                local_prompts = prompts_list[start_idx:end_idx]
                encoding = self.tokenizer.batch_encode_plus(local_prompts, padding=True, return_tensors="pt").to(
                    self.device)

            else:
                encoding = self.tokenizer.batch_encode_plus(prompts_list, padding=True, return_tensors="pt").to(
                    self.model.device)
            # encoding = self.tokenizer.batch_encode_plus(prompts_list, padding=True, return_tensors="pt").to(
            #     self.model.device)
            out = self.model.module.generate(encoding.input_ids, attention_mask=encoding.attention_mask, max_new_tokens=length,
                                      temperature=temperature)
            outs = self.tokenizer.batch_decode(out.tolist(), skip_special_tokens=True)

            return [out[len(self.tokenizer.decode(enc)):] for out, enc in zip(outs, encoding['input_ids'])]


import pytest

from textgrad.engine import get_engine

def test_ollama_engine():
    # Declare test constants
    OLLAMA_BASE_URL = 'http://localhost:11434/v1'
    MODEL_STRING = "test-model-string"

    # Initialise the engine
    engine = get_engine("ollama-" + MODEL_STRING)

    assert engine
    assert engine.model_string == MODEL_STRING
    assert engine.base_url == OLLAMA_BASE_URL


import os
import pandas as pd

class AbstractTask:
    def __init__(self, args, logger, data_dir):
        self.args = args
        self.logger = logger
        self.task = args.task
        self.subtask = args.subtask
        self.data_dir = data_dir
        
        self.data = {} # split_name: data

    def load_data(self):
        # ideally train.csv, dev.csv, test.csv in `self.data_dir`
        raise NotImplementedError

    def save_predictions(self, split="test", method=None):
        # rename the file as sth like test_pred_<method_name>.csv
        raise NotImplementedError

    def evaluate(self, labels, predictions):
        # return a list of per-example scores and an overall score
        raise NotImplementedError
    
    def get_data_split(self, split):
        if split not in self.data:
            raise Exception("data split {} does not exist.".format(split))
        else:
            return self.data[split]
    
    def __str__(self):
        s = "Task: {}; ".format(self.task)
        s += "Subtask: {}; ".format(self.subtask)
        s += "Data Dir: {}".format(self.data_dir)
        return s


"""
This codes runs a code generation pipeline for instance optimization.

The dataset used is the LeetCodeHardEval dataset. Part of the LeetCodeHard-Gym dataset.

Since we are evaluating using leetcode, the code has to run in two phases:

In the first phase, we parallel process the generation of the code using OpenAI's APIs.
In the second phase, we evaluate the code using the leetcode environment.

We need to do this because we cannot parallelize the leeetcode evaluation as it makes API calls for which
we have a rate limit.


"""
import copy
from dotenv import load_dotenv
import pandas as pd

load_dotenv()
import multiprocessing
from tqdm import tqdm
import textgrad
from textgrad.engine import get_engine
from textgrad import Variable
from textgrad.optimizer import TextualGradientDescent
from textgrad.tasks import load_instance_task
from prompts import CodeTestTimewithTests, SYSTEM_PROMPT_FOR_FIRST_CODE, CODE_INSTANCE_ROLE_DESCRIPTION
from evaluators.lt_eval import LeetCodeEvaluator
from evaluators.py_eval import PythonEvaluator

internal_evaluator = LeetCodeEvaluator()

def optimization_one_iteration(optimizer, instance_var, prompt, tests):
    """
    This is a single iteration of optimization
    :param optimizer:
    :param instance_var:
    :param prompt:
    :return:
    """
    pt = PythonEvaluator()
    tests = tests.split("\n")

    # Evaluate the code
    # state is True/False for each test
    state, test_string = pt.evaluate(instance_var.value, tests)

    optimizer.zero_grad()
    loss_fn = CodeTestTimewithTests(engine=ENGINE_API)

    test_time_loss = loss_fn(prompt, instance_var, test_string)

    test_time_loss.backward()
    optimizer.step()

    return state, test_string


def generate_starting_solution(prompt):
    """
    This is the first attempt at solving the problem.
    :param prompt:
    :return:
    """
    llm_first_code = ENGINE_API.generate(prompt, system_prompt=SYSTEM_PROMPT_FOR_FIRST_CODE)

    llm_first_code = llm_first_code.split("```python")[1].split("```")[0]
    return llm_first_code


def evaluation_and_optimization_pipeline(task_id, prompt, index, tests, MAX_ITERS):
    """
    :param prompt:
    :param index:
    :return:
    """
    print("Start of optimization pipeline", index)

    generated_programs = []
    gpt_4_first_code = generate_starting_solution(prompt)
    n_iter = 0
    generated_programs.append({"code": gpt_4_first_code,
                               "gradients": None,
                               "task_id": task_id,
                               "state": None,
                               "test_string": None,
                               "iteration": n_iter,
                               })

    instance_var = Variable(gpt_4_first_code, requires_grad=True,
                            role_description=CODE_INSTANCE_ROLE_DESCRIPTION)

    optimizer = TextualGradientDescent(engine=ENGINE_API,
                                       parameters=[instance_var],
                                       constraints=["Do not add asserts to the code",
                                                    "Code must contain imports"])

    pt = PythonEvaluator()
    # Evaluate the code
    # state is True/False for each test
    state, test_string = pt.evaluate(gpt_4_first_code, tests.split("\n"))

    generated_programs[-1]["state"] = state
    generated_programs[-1]["test_string"] = test_string
    # if all test passed we early stop

    for iter in range(1 + MAX_ITERS):
        state, test_string = optimization_one_iteration(optimizer, instance_var, prompt, tests)
        generated_programs[-1]["state"] = state
        generated_programs[-1]["test_string"] = test_string
        # if all test passed we early stop
        if ((iter != 0) and all(state)):
            break

        n_iter += 1
        generated_programs.append({"code": instance_var.value,
                                   "gradients": str(instance_var.gradients),
                                   "task_id": task_id,
                                   "state": None,
                                   "test_string": None,
                                   "iteration": n_iter,
                                   })


    print("End of optimization pipeline", index)

    # we get the first program (zero-shot) and the last program (after optimization, if it exists)
    return generated_programs


def evaluate_or_except(program):
    try:
        res = (internal_evaluator.check_if_in_cache_or_submit(program["task_id"], program["code"]))
        return res[0], res[1], res[2], res[3]
    except Exception as e:
        print(e)
        return False, -1, -1, -1


if __name__ == "__main__":

    SEEDS = [55]#, 91, 17, 21]#, 34]

    collection = {
        "task_id": [],
        "problem": [],
        "seed": [],
        "success": [],
        "test_cases_passed": [],
        "test_cases_total": [],
        "runtime": [],
        "local_tests": [],
        "local_test_state": [],
        "iteration_idx": []
    }


    for seed in SEEDS:
        instance_task = "LeetCodeHardEval"

        TEST_ENGINE = "gpt-4o"
        ENGINE_API = get_engine(engine_name=TEST_ENGINE, seed=seed)
        MAX_PROGRAMS_TO_OPTIMIZE = 39
        MULTIPROC_POOLS = min([20, MAX_PROGRAMS_TO_OPTIMIZE, multiprocessing.cpu_count()])
        MAX_ITERS = 4

        textgrad.set_backward_engine(ENGINE_API, override=True)

        dataset = load_instance_task(instance_task, ENGINE_API)

        code_dataset_examples = [(task_id, prompt, index, tests, MAX_ITERS)
                                 for index, (task_id, prompt, tests)
                                 in enumerate(dataset)][:MAX_PROGRAMS_TO_OPTIMIZE]

        with multiprocessing.Pool(MULTIPROC_POOLS) as pool:
            programs_and_gradients = pool.starmap(evaluation_and_optimization_pipeline, code_dataset_examples)

        all_programs = []
        zero_programs = []

        ctr = 0
        for list_of_programs in tqdm(programs_and_gradients):
            list_of_programs = [list_of_programs[0], list_of_programs[-1]]
            for iter, program in enumerate(list_of_programs):
                res = evaluate_or_except(program)
                program["evaluation"] = res[0]
                program["total_correct"] = res[1]
                program["total_tests"] = res[2]
                program["runtime"] = res[3]
                program["iteration_idx"] = iter
                all_programs.append(program)
                if ((program['iteration_idx'] == 0)):
                    ctr += 1
                    zero_programs.append(program)

        print("CounterZero", len(zero_programs))
        ctr = 0
        for program in all_programs:
            if program["iteration_idx"] == 0:
                ctr += 1
        print("CounterAll", ctr)
        
        for program in all_programs:
            collection["task_id"].append(program["task_id"])
            collection["problem"].append(program["code"])
            collection["seed"].append(seed)
            collection["iteration_idx"].append(program['iteration_idx'])
            collection["success"].append(program["evaluation"])
            collection["test_cases_passed"].append(program["total_correct"])
            collection["test_cases_total"].append(program["total_tests"])
            collection["runtime"].append(program["runtime"])
            collection["local_tests"].append(program["test_string"])
            collection["local_test_state"].append(program["state"])
            

        program_df = pd.DataFrame(collection)

        program_df.to_csv(f"../results/code_optimization/textgrad/leetcode_supervised_{TEST_ENGINE}.csv", index=False)




from .ape_trainer import APETrainer
from .apo_trainer import APOTrainer
from .default_trainer import DefaultTrainer
from .pe2_trainer import PE2Trainer

TRAINER2CLASS = {
    "ape": APETrainer,
    "apo": APOTrainer,
    "default": DefaultTrainer,
    "pe2": PE2Trainer
}

def Trainer2Class(trainer):
    if trainer not in TRAINER2CLASS:
        raise NotImplementedError
    else:
        return TRAINER2CLASS[trainer]


from vllm import LLM, SamplingParams

# Sample prompts.
prompts = [
    "Hello, my name is",
    "The president of the United States is",
    "The capital of France is",
    "The future of AI is",
]
# Create a sampling params object.
sampling_params = SamplingParams(temperature=0.0)

# Create an LLM.
# Test single-GPU inference
llm = LLM(model="facebook/opt-125m")

# Generate texts from the prompts. The output is a list of RequestOutput objects
# that contain the prompt, generated text, and other information.
outputs = llm.generate(prompts, sampling_params)
# Print the outputs.
for output in outputs:
    prompt = output.prompt
    generated_text = output.outputs[0].text
    print(f"Prompt: {prompt!r}, Generated text: {generated_text!r}")

## Expected Output:
### Prompt: 'Hello, my name is', Generated text: ' J.C. and I am a student at the University of California, Berkeley'                                                                           
### Prompt: 'The president of the United States is', Generated text: " not a racist. He is a racist.\nHe's a racist because he"                                                                 
### Prompt: 'The capital of France is', Generated text: ' the capital of the French Republic.\n\nThe capital of France is the capital'                                                          
### Prompt: 'The future of AI is', Generated text: ' in the hands of the people.\n\nThe future of AI is in the'  


import concurrent
import csv
import importlib
import os
from collections import defaultdict
from typing import List

from .core.TextGrad.textgrad_ollm import textgrad as tg
from .core.TextGrad.textgrad_ollm.textgrad.tasks import load_task
from .core.TextGrad.textgrad_ollm.textgrad.engine.local_model_openai_api import ChatExternalClient
from .core.pe2.cli import ape_apo_pe2_optimizer
from .dataloader import GreaterDataloader
from .models import model_supported
from .utils import ape_apo_pe2_args, clean_string, textgrad_args

import torch
import numpy as np
import pandas as pd
from torch.nn import functional as F
from tqdm import tqdm
from transformers import AutoModelForCausalLM, AutoTokenizer


class BaseOptimizer:
    def __init__(self, trainer: str, optimize_config: dict = {}):
        self.args = ape_apo_pe2_args(trainer)
        self.args.task_model = optimize_config.get("task_model", "openai_gpt35_turbo_instruct")
        self.args.optim_model = optimize_config.get("optim_model", "openai_gpt4_turbo")


    def write_data(self, inputs: GreaterDataloader, p_init: str) -> None:
        os.makedirs(self.args.data_dir, exist_ok=True)

        for filename in ["train", "test", "dev"]:
            with open(os.path.join(self.args.data_dir, f"{filename}.csv"), "w") as f:
                writer = csv.writer(f)
                writer.writerow(["", "input", "label"])

                for i, input in enumerate(inputs):
                    writer.writerow([i, input["question"], input["answer"]])
        
        with open(os.path.join(self.args.data_dir, "prompt.md"), "w") as f:
            f.write(p_init)
        
        return


    def optimize(self, inputs: GreaterDataloader, p_init: str) -> str:
        self.write_data(inputs, p_init)

        p_star = ape_apo_pe2_optimizer(self.args)

        return p_star


class ApeOptimizer(BaseOptimizer):
    def __init__(self, optimize_config: dict = {}):
        super().__init__("ape", optimize_config)


class ApoOptimizer(BaseOptimizer):
    def __init__(self, optimize_config: dict = {}):
        super().__init__("apo", optimize_config)


class GreaterOptimizer:
    def __init__(
            self, model: AutoModelForCausalLM, tokenizer: AutoTokenizer,
            optimize_config: dict = {}, *args, **kwargs
        ):
        self.optimize_config = optimize_config
        self._init_agents(model, tokenizer)
    

    def _init_agents(self, model: AutoModelForCausalLM, tokenizer: AutoTokenizer):
        supported, model_name = model_supported(model)
        assert supported, f"Model: {model} is not supported"

        model_class = getattr(importlib.import_module("src.greaterprompt.models"), model_name)
        self.client = model_class(model, tokenizer)

        for param in self.client.model.parameters():
            param.requires_grad = False
        self.client.model.get_input_embeddings().weight.requires_grad = True


    def encode_input(self, batch_inputs: List[dict], p_extractor: str) -> torch.Tensor:
        q_tokens, p_tokens, y_tokens = [], [], []
        for input in batch_inputs:
            q_token = self.client.tokenizer.encode(input["question"].strip() + " ?", return_tensors="pt")
            p_token = self.client.tokenizer.encode(" " + input["prompt"], return_tensors="pt")
            if self.client.model.config.model_type == "llama":
                y_token = self.client.tokenizer.encode(" " + input["answer"], return_tensors="pt")
            elif self.client.model.config.model_type == "gemma2":
                y_token = self.client.tokenizer.encode(input["answer"], return_tensors="pt")
            q_tokens.append(q_token)
            p_tokens.append(p_token[:, 1:])
            y_tokens.append(y_token[:, 1:])
        
        p_extr_token = self.client.tokenizer.encode(p_extractor, return_tensors="pt")
        p_extr_token = p_extr_token[:, 1:]

        return (
            [t.to(self.client.device, dtype=torch.long) for t in q_tokens],
            [t.to(self.client.device, dtype=torch.long) for t in p_tokens],
            p_extr_token.to(self.client.device, dtype=torch.long),
            [t.to(self.client.device, dtype=torch.long) for t in y_tokens]
        )
    

    def get_y_hat(self, input: torch.Tensor, y_tokens: torch.Tensor) -> torch.Tensor:
        y_hat = None

        with torch.enable_grad():
            generate_config = self.optimize_config["generate_config"]
            for i in range(len(y_tokens)):
                logits = self.client.get_logits(input, generate_config)[:, -1, :]
                if y_hat is None:
                    y_hat = logits
                else:
                    y_hat = torch.cat([y_hat, logits], dim=0)
                probs = F.softmax(logits, dim=1)
                next_token_id = torch.argmax(probs, dim=1).reshape(1, -1)

                # ----------------- only for gemma2 ----------------- #
                # TODO: hardcode here, if gemma2 generate \n, keep generating until it is not
                if self.client.model.config.model_type == "gemma2":
                    while self.client.tokenizer.decode(next_token_id[0], skip_special_tokens=True) in ["\n", "\n\n"]:
                        input = torch.cat([input, next_token_id], dim=1)
                        logits = self.client.get_logits(input, generate_config)[:, -1, :]
                        probs = F.softmax(logits, dim=1)
                        next_token_id = torch.argmax(probs, dim=1).reshape(1, -1)
                # ----------------- only for gemma2 ----------------- #

                input = torch.cat([input, next_token_id], dim=1)
                del logits
                if i % 5 == 0:
                    torch.cuda.empty_cache()

        return y_hat
    

    def perplexity_loss(self, q_tokens: torch.Tensor, p_tokens: torch.Tensor) -> torch.Tensor:
        loss = torch.tensor(0.0, device=p_tokens.device)
        
        for i in range(1, p_tokens.size(1)):
            self.client.model.zero_grad()
            input_ids = torch.cat([q_tokens, p_tokens[:, :i]], dim=1)
            logits = self.client.get_logits(input_ids, self.optimize_config["generate_config"])[-1, -1, :]
            log_probs = F.log_softmax(logits, dim=-1)
            loss += log_probs[p_tokens[0, i]]
        
        loss = torch.exp(-1 * loss / p_tokens.size(1))
        return loss


    def calculate_loss(self, q_tokens: torch.Tensor, p_tokens: torch.Tensor, y_hat: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        self.client.model.zero_grad()

        loss_function = self.optimize_config.get("loss_function", F.cross_entropy)
        if self.optimize_config.get("perplexity_loss", False):
            perpl_lambda = self.optimize_config.get("perplexity_lambda", 0.2)
        else:
            perpl_lambda = 0
        raw_loss = loss_function(y_hat, y[0, :]).mean()
        perpl_loss = perpl_lambda * self.perplexity_loss(q_tokens, p_tokens)
        loss = raw_loss + perpl_loss
        loss.backward()

        return raw_loss


    def get_gradients(self, q_tokens: torch.Tensor, p_tokens: torch.Tensor, y_tokens: torch.Tensor, y_hat: torch.Tensor) -> List[torch.Tensor]:
        loss = self.calculate_loss(q_tokens, p_tokens, y_hat, y_tokens)
        embedding_layer = self.client.model.get_input_embeddings()
        embedding_grad = embedding_layer.weight.grad.detach().clone()

        return embedding_grad, loss


    def get_candidates(self, input: torch.Tensor) -> List[int]:
        with torch.inference_mode():
            candidates = self.client.get_candidates(input, self.optimize_config)

        return candidates
    

    def get_reasoning(self, input: torch.Tensor) -> str:
        generate_config = self.optimize_config.get("generate_config", {})
        with torch.inference_mode():
            response = self.client.generate(input, generate_config)

        return response
    
    
    def get_p_i_star(self, gradients: torch.Tensor, candidates: List[int]) -> int:
        p_i_star, p_i_star_grad = None, float("-inf")

        for candidate in candidates:
            token_grad = -1 * gradients[candidate, :].mean().item()
            if token_grad > p_i_star_grad:
                p_i_star = candidate
                p_i_star_grad = token_grad

        return p_i_star


    def optimize(self, inputs: GreaterDataloader, p_extractor: str, rounds: int) -> dict[str, List[str]]:
        outputs = defaultdict(list)

        intersect_q = self.optimize_config.get("intersect_q", 1)
        batch_inputs = [inputs[i:i+intersect_q] for i in range(0, len(inputs), intersect_q)]

        for i, batch in enumerate(batch_inputs):
            question_tokens, p_tokens, p_extr_tokens, y_tokens = self.encode_input(batch, p_extractor)
            assert all(p_token.size(1) >= 2 for p_token in p_tokens), "Init prompt should be at least 2 tokens"
            p_stars = defaultdict(list)
            idx = 1
            truncated = [False] * len(batch)

            for j in tqdm(range(rounds), desc=f"Optimizing {i + 1} / {len(batch_inputs)}"):
                torch.cuda.empty_cache()
                # get candidates for p_i by using x + p_0 ... p_i-1 for each p in the batch
                candidates_tmp = []

                for k, p in enumerate(p_tokens):
                    if idx >= p.size(1): continue
                    p_token_i = p[:, idx]
                    input_ids = torch.cat([question_tokens[k], p[:, :idx]], dim=1)
                    p_candidates = set(self.get_candidates(input_ids))
                    # avoid adding dummy token to candidates
                    eos_token_id = self.client.tokenizer.eos_token_id
                    p_candidates.update([int(p_token_i[0])] if int(p_token_i[0]) != eos_token_id else [])
                    candidates_tmp.append(p_candidates)
                
                candidates = set(candidates_tmp[0])
                for c in candidates_tmp[1:]:
                    candidates = candidates.intersection(c)
                if not candidates:
                    candidates = set(t for c in candidates_tmp for t in c)
                
                # use intersection candidates to optimize each p in the batch
                for k, p in enumerate(p_tokens):
                    if idx >= p.size(1): continue
                    # use x + p to get reasoning chain r
                    input_ids = torch.cat([question_tokens[k], p], dim=1)
                    reasoning_chain = self.get_reasoning(input_ids)
                    r_tokens = self.client.tokenizer.encode(reasoning_chain, return_tensors="pt")
                    r_tokens = r_tokens.to(self.client.device)

                    # use x + p + r + p_extractor to get logits of y_hat(x and p is already included in r)
                    input_ids = torch.cat([r_tokens, p_extr_tokens], dim=1)
                    y_hat = self.get_y_hat(input_ids, y_tokens[k])
                    gradients, loss = self.get_gradients(question_tokens[k], p[:, :idx], y_tokens[k], y_hat)

                    # calculate gradient for each candidate to get p_i_star
                    p_i_star = self.get_p_i_star(gradients, candidates)
                    p[:, idx] = p_i_star

                    # if p_i_star is a period, truncate the prompt and star from the beginning
                    p_i_star_token = self.client.tokenizer.decode(p_i_star)
                    if p_i_star_token.strip() in ".?!:":
                        p_tokens[k] = p[:, :idx + 1]
                        truncated[k] = True
                        decoded_p_star = self.client.tokenizer.decode(p_tokens[k][0, :], skip_special_tokens=True)
                        p_stars[batch[k]["question"]].append((repr(decoded_p_star.strip()), loss.item()))
                    # add a dummy token to the end of the prompt for next round of candidate generation
                    elif p_i_star_token.strip() != "." and idx == len(p[0, :]) - 1:
                        eos_token_id = self.client.tokenizer.eos_token_id
                        dummy_token = torch.tensor([[eos_token_id]], device=p.device)
                        p_tokens[k] = torch.cat([p, dummy_token], dim=1)

                if all(t for t in truncated):
                    idx = 1
                    truncated = [False] * len(batch)
                else:
                    idx = (idx + 1) % max(p.size(1) for p in p_tokens)
            
            # save and filter the optimized prompts
            outputs.update(p_stars)
            for k, p in enumerate(p_tokens):
                outputs[batch[k]["question"]].append((
                    repr(self.client.tokenizer.decode(p[0, :], skip_special_tokens=True).strip()),
                    loss.item()
                ))
            for question, prompts in outputs.items():
                cleaned_prompts = clean_string(prompts)
                if self.optimize_config.get("filter", False):
                    outputs[question] = self.client.filter(cleaned_prompts)

            del input_ids, reasoning_chain, r_tokens, y_hat
            torch.cuda.empty_cache()

        return outputs


    def optimize_streamlit(self, inputs: GreaterDataloader, p_extractor: str, rounds: int, callback=None) -> dict[str, List[str]]:
        outputs = defaultdict(list)

        intersect_q = self.optimize_config.get("intersect_q", 1)
        batch_inputs = [inputs[i:i+intersect_q] for i in range(0, len(inputs), intersect_q)]

        total_batches = len(batch_inputs)

        for i, batch in enumerate(batch_inputs):
            question_tokens, p_tokens, p_extr_tokens, y_tokens = self.encode_input(batch, p_extractor)
            assert all(p_token.size(1) >= 2 for p_token in p_tokens), "Init prompt should be at least 2 tokens"
            p_stars = defaultdict(list)
            idx = 1
            truncated = [False] * len(batch)

            for j in range(rounds):
                if callback:
                    batch_progress = (i / total_batches)
                    round_progress = (j / rounds) / total_batches
                    total_progress = batch_progress + round_progress
                    
                    current_loss = None
                    if p_stars:
                        for question in p_stars:
                            if p_stars[question]:
                                _, loss_value = p_stars[question][-1]
                                if current_loss is None or loss_value < current_loss:
                                    current_loss = loss_value
                    
                    status_info = {
                        "batch": i + 1,
                        "total_batches": total_batches,
                        "round": j + 1,
                        "total_rounds": rounds,
                        "loss": current_loss
                    }
                    
                    callback(total_progress, status_info)
                
                torch.cuda.empty_cache()
                # get candidates for p_i by using x + p_0 ... p_i-1 for each p in the batch
                candidates_tmp = []

                for k, p in enumerate(p_tokens):
                    if idx >= p.size(1): continue
                    p_token_i = p[:, idx]
                    input_ids = torch.cat([question_tokens[k], p[:, :idx]], dim=1)
                    p_candidates = set(self.get_candidates(input_ids))
                    # avoid adding dummy token to candidates
                    eos_token_id = self.client.tokenizer.eos_token_id
                    p_candidates.update([int(p_token_i[0])] if int(p_token_i[0]) != eos_token_id else [])
                    candidates_tmp.append(p_candidates)
                
                candidates = set(candidates_tmp[0])
                for c in candidates_tmp[1:]:
                    candidates = candidates.intersection(c)
                if not candidates:
                    candidates = set(t for c in candidates_tmp for t in c)
                
                # use intersection candidates to optimize each p in the batch
                for k, p in enumerate(p_tokens):
                    if idx >= p.size(1): continue
                    # use x + p to get reasoning chain r
                    input_ids = torch.cat([question_tokens[k], p], dim=1)
                    reasoning_chain = self.get_reasoning(input_ids)
                    r_tokens = self.client.tokenizer.encode(reasoning_chain, return_tensors="pt")
                    r_tokens = r_tokens.to(self.client.device)

                    # use x + p + r + p_extractor to get logits of y_hat(x and p is already included in r)
                    input_ids = torch.cat([r_tokens, p_extr_tokens], dim=1)
                    y_hat_probs = self.get_pred_probs(input_ids, y_tokens[k])
                    gradients, loss = self.get_gradients(question_tokens[k], p[:, :idx], y_tokens[k], y_hat_probs)

                    # calculate gradient for each candidate to get p_i_star
                    p_i_star = self.get_p_i_star(gradients, candidates)
                    p[:, idx] = p_i_star

                    # if p_i_star is a period, truncate the prompt and star from the beginning
                    p_i_star_token = self.client.tokenizer.decode(p_i_star)
                    if p_i_star_token.strip() in ".?!:":
                        p_tokens[k] = p[:, :idx + 1]
                        truncated[k] = True
                        decoded_p_star = self.client.tokenizer.decode(p_tokens[k][0, :], skip_special_tokens=True)
                        p_stars[batch[k]["question"]].append((repr(decoded_p_star.strip()), loss.item()))
                    # add a dummy token to the end of the prompt for next round of candidate generation
                    elif p_i_star_token.strip() != "." and idx == len(p[0, :]) - 1:
                        eos_token_id = self.client.tokenizer.eos_token_id
                        dummy_token = torch.tensor([[eos_token_id]], device=p.device)
                        p_tokens[k] = torch.cat([p, dummy_token], dim=1)

                if all(t for t in truncated):
                    idx = 1
                    truncated = [False] * len(batch)
                else:
                    idx = (idx + 1) % max(p.size(1) for p in p_tokens)
            
            # save and filter the optimized prompts
            outputs.update(p_stars)
            for k, p in enumerate(p_tokens):
                outputs[batch[k]["question"]].append((
                    repr(self.client.tokenizer.decode(p[0, :], skip_special_tokens=True).strip()),
                    loss.item()
                ))
            for question, prompts in outputs.items():
                cleaned_prompts = clean_string(prompts)
                if self.optimize_config.get("filter", False):
                    outputs[question] = self.client.filter(cleaned_prompts)

            del input_ids, reasoning_chain, r_tokens, y_hat_probs
            torch.cuda.empty_cache()

        if callback:
            callback(1.0, {"status": "complete"})

        return outputs


class Pe2Optimizer(BaseOptimizer):
    def __init__(self, optimize_config: dict = {}):
        super().__init__("pe2", optimize_config)


class TextGradOptimizer:
    def __init__(self, optimize_config: dict = {}):
        self.args = textgrad_args()
        self.args.evaluation_engine = optimize_config.get("evaluation_engine", "meta-llama/Meta-Llama-3-8B-Instruct")
        self.args.test_engine = optimize_config.get("test_engine", "meta-llama/Meta-Llama-3-8B-Instruct")
        self.args.device = optimize_config.get("device", "cpu")

        evaluation_engine = ChatExternalClient(client=None, model_string=self.args.evaluation_engine, device=self.args.device)
        if self.args.evaluation_engine != self.args.test_engine:
            test_engine = ChatExternalClient(client=None, model_string=self.args.test_engine, device=self.args.device)
        else:
            test_engine = evaluation_engine
        self.llm_api_eval = evaluation_engine
        self.llm_api_test = test_engine

    
    def load_data(self, p_init:str, dataloader:GreaterDataloader):
        train_set, val_set, test_set, eval_fn, csv_path = load_task(
            p_init, self.args.task, evaluation_api=self.args.evaluation_engine
        )
        for filename in ["train", "test", "val"]:
            with open(os.path.join(csv_path, f"{filename}.csv"), "w") as f:
                writer = csv.writer(f)
                writer.writerow(["", "x", "y"])
                for i, input in enumerate(dataloader):
                    writer.writerow([i, input["question"], input["answer"]])
            
            if filename == "train":
                train_set.data = pd.read_csv(os.path.join(csv_path, f"{filename}.csv"), index_col=0)
            elif filename == "test":
                test_set.data = pd.read_csv(os.path.join(csv_path, f"{filename}.csv"), index_col=0)
            elif filename == "val":
                val_set.data = pd.read_csv(os.path.join(csv_path, f"{filename}.csv"), index_col=0)

        return train_set, val_set, test_set, eval_fn


    def eval_sample(self, item, eval_fn, model):
        x, y = item
        y = str(y)
        x = tg.Variable(x, requires_grad=False, role_description="query to the language model")
        y = tg.Variable(str(y), requires_grad=False, role_description="correct answer for the query")
        response = model(x)
        try:
            eval_output_variable = eval_fn(inputs=dict(prediction=response, ground_truth_answer=y))
            return int(eval_output_variable.value)
        except:
            eval_output_variable = eval_fn([x, y, response])
            eval_output_parsed = eval_fn.parse_output(eval_output_variable)
            return int(eval_output_parsed)


    def eval_dataset(self, test_set, eval_fn, model, max_samples: int=None):
        if max_samples is None:
            max_samples = len(test_set)
        accuracy_list = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.args.num_threads) as executor:
            futures = []
            for _, sample in enumerate(test_set):
                
                future = executor.submit(self.eval_sample, sample, eval_fn, model)
                futures.append(future)
                if len(futures) >= max_samples:
                    break
            tqdm_loader = tqdm(concurrent.futures.as_completed(futures), total=len(futures), position=0)
            for future in tqdm_loader:
                acc_item = future.result()
                accuracy_list.append(acc_item)
                tqdm_loader.set_description(f"Proxy Accuracy: {np.mean(accuracy_list)}")
        return accuracy_list
    

    def run_validation_revert(self, system_prompt: tg.Variable, results, model, eval_fn, val_set):
        val_performance = np.mean(self.eval_dataset(val_set, eval_fn, model))
        previous_performance = np.mean(results["validation_acc"][-1])
        previous_prompt = results["prompt"][-1]
        
        if val_performance < previous_performance:
            system_prompt.set_value(previous_prompt)
            val_performance = previous_performance

        results["validation_acc"].append(val_performance)


    def optimize(self, dataloader: GreaterDataloader, p_init: str):
        tg.set_backward_engine(self.llm_api_eval, override=True)

        train_set, val_set, test_set, eval_fn = self.load_data(p_init, dataloader)
        STARTING_SYSTEM_PROMPT = train_set.get_task_description()

        train_loader = tg.tasks.DataLoader(train_set, batch_size=self.args.batch_size, shuffle=True)
        system_prompt = tg.Variable(
            STARTING_SYSTEM_PROMPT, requires_grad=True, role_description="system prompt to the language model"
        )
        model_evaluation = tg.BlackboxLLM(self.llm_api_eval, system_prompt)

        if not self.args.do_not_run_larger_model:
            reference = np.mean(self.eval_dataset(test_set, eval_fn, model_evaluation))


        system_prompt = tg.Variable(
            STARTING_SYSTEM_PROMPT, requires_grad=True,
            role_description="structured system prompt to a somewhat capable language model \
                  that specifies the behavior and strategies for the QA task"
        )
        model = tg.BlackboxLLM(self.llm_api_test, system_prompt)

        optimizer = tg.TextualGradientDescent(engine=self.llm_api_eval, parameters=[system_prompt])

        results = {"test_acc": [], "prompt": [], "validation_acc": []}
        results["test_acc"].append(self.eval_dataset(test_set, eval_fn, model))
        results["validation_acc"].append(self.eval_dataset(val_set, eval_fn, model))
        results["prompt"].append(system_prompt.get_value())

        for epoch in range(self.args.max_epochs):
            for steps, (batch_x, batch_y) in enumerate((pbar := tqdm(train_loader, position=0))):
                pbar.set_description(f"Training step {steps}. Epoch {epoch}")
                optimizer.zero_grad()
                losses = []
                for (x, y) in zip(batch_x, batch_y):
                    y = str(y)
                    x = tg.Variable(x, requires_grad=False, role_description="query to the language model")
                    y = tg.Variable(y, requires_grad=False, role_description="correct answer for the query")
                    response = model(x)
                    try:
                        eval_output_variable = eval_fn(inputs=dict(prediction=response, ground_truth_answer=y))
                    except:
                        eval_output_variable = eval_fn([x, y, response])
                    losses.append(eval_output_variable)
                total_loss = tg.sum(losses)
                total_loss.backward()
                optimizer.step()
                if self.args.run_validation:
                    self.run_validation_revert(system_prompt, results, model, eval_fn, val_set)
                test_acc = self.eval_dataset(test_set, eval_fn, model)
                results["test_acc"].append(test_acc)
                results["prompt"].append(system_prompt.get_value())
                if steps == 3:
                    break
        
        p_star = results["prompt"][-1]

        return p_star
